import logging
import logging.config
import logging.handlers
from path import Path


def init_logger(
    level="INFO", log_dir="./", log_name="main_logger", filename="main.log"
):
    logger = logging.getLogger(log_name)

    fh = logging.handlers.RotatingFileHandler(
        Path(log_dir) / filename, "w", 20 * 1024 * 1024, 5
    )
    formatter = logging.Formatter(
        "%(asctime)s %(levelname)5s - %(name)s "
        "[%(filename)s line %(lineno)d] - %(message)s",
        datefmt="%m-%d %H:%M:%S",
    )
    fh.setFormatter(formatter)
    logger.addHandler(fh)

    # logging to screen
    fh = logging.StreamHandler()
    formatter = logging.Formatter(
        "[%(levelname)s] %(message)s",
    )
    fh.setFormatter(formatter)
    logger.addHandler(fh)

    logger.setLevel(level)
    logger.info("Start training")
    return logger



import torch
from neurovc.contrib.flowmag.flow_models.gmflow.position import PositionEmbeddingSine


def split_feature(
    feature,
    num_splits=2,
    channel_last=False,
):
    if channel_last:  # [B, H, W, C]
        b, h, w, c = feature.size()
        assert h % num_splits == 0 and w % num_splits == 0

        b_new = b * num_splits * num_splits
        h_new = h // num_splits
        w_new = w // num_splits

        feature = (
            feature.view(b, num_splits, h // num_splits, num_splits, w // num_splits, c)
            .permute(0, 1, 3, 2, 4, 5)
            .reshape(b_new, h_new, w_new, c)
        )  # [B*K*K, H/K, W/K, C]
    else:  # [B, C, H, W]
        b, c, h, w = feature.size()
        assert h % num_splits == 0 and w % num_splits == 0

        b_new = b * num_splits * num_splits
        h_new = h // num_splits
        w_new = w // num_splits

        feature = (
            feature.view(b, c, num_splits, h // num_splits, num_splits, w // num_splits)
            .permute(0, 2, 4, 1, 3, 5)
            .reshape(b_new, c, h_new, w_new)
        )  # [B*K*K, C, H/K, W/K]

    return feature


def merge_splits(
    splits,
    num_splits=2,
    channel_last=False,
):
    if channel_last:  # [B*K*K, H/K, W/K, C]
        b, h, w, c = splits.size()
        new_b = b // num_splits // num_splits

        splits = splits.view(new_b, num_splits, num_splits, h, w, c)
        merge = (
            splits.permute(0, 1, 3, 2, 4, 5)
            .contiguous()
            .view(new_b, num_splits * h, num_splits * w, c)
        )  # [B, H, W, C]
    else:  # [B*K*K, C, H/K, W/K]
        b, c, h, w = splits.size()
        new_b = b // num_splits // num_splits

        splits = splits.view(new_b, num_splits, num_splits, c, h, w)
        merge = (
            splits.permute(0, 3, 1, 4, 2, 5)
            .contiguous()
            .view(new_b, c, num_splits * h, num_splits * w)
        )  # [B, C, H, W]

    return merge


def normalize_img(img0, img1):
    # loaded images are in [0, 255]
    # normalize by ImageNet mean and std
    mean = torch.tensor([0.485, 0.456, 0.406]).view(1, 3, 1, 1).to(img1.device)
    std = torch.tensor([0.229, 0.224, 0.225]).view(1, 3, 1, 1).to(img1.device)
    img0 = (img0 / 255.0 - mean) / std
    img1 = (img1 / 255.0 - mean) / std

    return img0, img1


def feature_add_position(feature0, feature1, attn_splits, feature_channels):
    pos_enc = PositionEmbeddingSine(num_pos_feats=feature_channels // 2)

    if attn_splits > 1:  # add position in splited window
        feature0_splits = split_feature(feature0, num_splits=attn_splits)
        feature1_splits = split_feature(feature1, num_splits=attn_splits)

        position = pos_enc(feature0_splits)

        feature0_splits = feature0_splits + position
        feature1_splits = feature1_splits + position

        feature0 = merge_splits(feature0_splits, num_splits=attn_splits)
        feature1 = merge_splits(feature1_splits, num_splits=attn_splits)
    else:
        position = pos_enc(feature0)

        feature0 = feature0 + position
        feature1 = feature1 + position

    return feature0, feature1



import importlib
import importlib.util
from pathlib import Path
from types import ModuleType

import cv2
import numpy as np

from neurovc.util import normalize_color

HAS_YOLOV5_FACE = importlib.util.find_spec("yolov5_face") is not None
HAS_GDOWN = importlib.util.find_spec("gdown") is not None


class OptionalDependencyError(ImportError):
    pass


def require(module: str, *, extra: str, purpose: str) -> ModuleType:
    root = module.split(".", 1)[0]
    root_spec = importlib.util.find_spec(root)
    if root_spec is None:
        raise OptionalDependencyError(
            f"Missing optional dependency '{root}' required for {purpose}. "
            f"Install via `pip install neurovc[{extra}]`."
        )

    try:
        return importlib.import_module(module)
    except ModuleNotFoundError as exc:
        raise OptionalDependencyError(
            f"Dependency '{root}' is installed, but importing '{module}' failed "
            f"(likely incompatible/partial install). "
            f"Try `pip install -U --force-reinstall {root}`. "
            f"Original error: {exc.__class__.__name__}: {exc}"
        ) from exc
    except Exception as exc:
        raise OptionalDependencyError(
            f"Dependency '{module}' was found but failed to import (broken binary/env). "
            f"Original error: {exc.__class__.__name__}: {exc}"
        ) from exc


class _ModelDownloader:
    def __init__(
        self,
        model_name: str = "YOLOv5n-Face.modern",
        save_dir: Path | str | None = None,
    ):
        self.model_name = model_name
        if self.model_name not in _file_id_map:
            available = sorted(k for k in _file_id_map if k != "license")
            raise ValueError(
                f"Model name '{model_name}' is not valid. Available: {available}"
            )

        base_dir = (
            Path(save_dir) if save_dir is not None else Path("~/.neurovc/models/tfw")
        )
        self.save_dir = base_dir.expanduser()
        self.model_path = self.save_dir / _file_targets.get(
            model_name, f"{model_name}.pt"
        )
        try:
            self.license_id = _file_id_map["license"]
        except KeyError as exc:
            raise ValueError("License file id is missing from _file_id_map.") from exc

    @staticmethod
    def _download_file(file_id: str, output_path: Path) -> Path:
        if output_path.exists():
            return output_path

        gdown = require("gdown", extra="landmark", purpose="TFW model download")

        url = f"https://drive.google.com/uc?export=download&id={file_id}"
        gdown.download(url, str(output_path), quiet=False)
        return output_path

    def download_model(self) -> Path:
        self.save_dir.mkdir(parents=True, exist_ok=True)
        model_id = _file_id_map[self.model_name]
        model_path = self._download_file(model_id, self.model_path)
        self._download_file(self.license_id, self.save_dir / _file_targets["license"])
        return model_path


def _prepare_model(model_name: str = "YOLOv5n-Face.modern") -> Path:
    downloader = _ModelDownloader(model_name)
    return downloader.download_model()


class LandmarkWrapper:
    def process(self, img):
        return self.get_landmarks(img), None


class TFWLandmarker(LandmarkWrapper):
    def __init__(self, model_name="YOLOv5n-Face.modern", device=None):
        yf = require(
            "yolov5_face.detect_face", extra="landmark", purpose="TFWLandmarker"
        )

        self._yf = yf

        self.device = device
        model_path = _prepare_model(model_name)
        self.model = (
            yf.load_model(model_path, self.device)
            if self.device
            else yf.load_model(model_path)
        )

    def detect(self, img):
        img = normalize_color(img, color_map=cv2.COLORMAP_BONE)
        img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
        return (
            self._yf.detect_landmarks(self.model, img, self.device)
            if self.device
            else self._yf.detect_landmarks(self.model, img)
        )

    def get_landmarks(self, img):
        results = self.detect(img)
        if len(results) == 0:
            return np.full((5, 2), -1)
        lm = results[0]["landmarks"]
        lm = np.array(lm).reshape((-1, 2))
        return lm


_file_id_map = {
    # "YOLOv5n": "1PLUq7WbOWS7Ve2VKW7_WBkC3Uksje8Fx",
    # "YOLOv5n6": "1wV9t5uH_eiy7WaHdQdWnbeEIijuDAdKI",
    # "YOLOv5s": "1IdsdR1-qUeRo5EKQJzGQmRDi2SrMXJG5",
    # "YOLOv5s6": "1YZX3t7cSPnWWoic7oJo86ljBQgE5PPb2",
    # "YOLOv5n-Face": "1vXk9P3CfhUtRBGI44SqWbuiTJ7rAI4hP",
    "YOLOv5n-Face.modern": "14MpWOt-LEWM1w1XxMCngXKAYbt1toqrA",
    "license": "13jydQUIgVjK4XDdPXhVweDv5t1_n6iJy",
}

_file_targets = {
    "YOLOv5n-Face.modern": "YOLOv5n-Face.modern.pt",
    "license": "license.txt",
}


__all__ = [
    "HAS_YOLOV5_FACE",
    "HAS_GDOWN",
    "OptionalDependencyError",
    "require",
    "TFWLandmarker",
    "LandmarkWrapper",
]



import numbers
import random
import numpy as np
# from scipy.misc import imresize


def get_co_transforms(aug_args):
    transforms = []
    if aug_args.crop:
        transforms.append(RandomCrop(aug_args.para_crop))
    if aug_args.hflip:
        transforms.append(RandomHorizontalFlip())
    if aug_args.swap:
        transforms.append(RandomSwap())
    return Compose(transforms)


class Compose(object):
    def __init__(self, co_transforms):
        self.co_transforms = co_transforms

    def __call__(self, input, target):
        for t in self.co_transforms:
            input, target = t(input, target)
        return input, target


class RandomCrop(object):
    """Crops the given PIL.Image at a random location to have a region of
    the given size. size can be a tuple (target_height, target_width)
    or an integer, in which case the target will be of a square shape (size, size)
    """

    def __init__(self, size):
        if isinstance(size, numbers.Number):
            self.size = (int(size), int(size))
        else:
            self.size = size

    def __call__(self, inputs, target):
        h, w, _ = inputs[0].shape
        th, tw = self.size
        if w == tw and h == th:
            return inputs, target

        x1 = random.randint(0, w - tw)
        y1 = random.randint(0, h - th)
        inputs = [img[y1 : y1 + th, x1 : x1 + tw] for img in inputs]
        if "mask" in target:
            target["mask"] = target["mask"][y1 : y1 + th, x1 : x1 + tw]
        if "flow" in target:
            target["flow"] = target["flow"][y1 : y1 + th, x1 : x1 + tw]
        return inputs, target


class RandomSwap(object):
    def __call__(self, inputs, target):
        if random.random() < 0.5:
            inputs = inputs[::-1]
            if "mask" in target:
                target["mask"] = target["mask"][::-1]
            if "flow" in target:
                raise NotImplementedError("swap cannot apply to flow")
        return inputs, target


class RandomHorizontalFlip(object):
    """Randomly horizontally flips the given PIL.Image with a probability of 0.5"""

    def __call__(self, inputs, target):
        if random.random() < 0.5:
            inputs = [np.copy(np.fliplr(im)) for im in inputs]
            if "mask" in target:
                target["mask"] = [np.copy(np.fliplr(mask)) for mask in target["mask"]]
            if "flow" in target:
                for i, flo in enumerate(target["flow"]):
                    flo = np.copy(np.fliplr(flo))
                    flo[:, :, 0] *= -1
                    target["flow"][i] = flo
        return inputs, target



"""Utility helpers for FlowMag."""

from neurovc.contrib.flowmag_util.models import (
    download_all_flowmag_models,
    download_flowmag_model,
    flowmag_model_downloader,
)
from neurovc.contrib.flowmag_util.online import (
    FlowMagOnline,
    FlowMagTTA,
    OnlineMagnifier,
    TTAStep,
    default_alpha_policy,
    load_flowmag_model,
    unwrap,
)

__all__ = [
    "download_all_flowmag_models",
    "download_flowmag_model",
    "flowmag_model_downloader",
    "FlowMagOnline",
    "FlowMagTTA",
    "OnlineMagnifier",
    "TTAStep",
    "default_alpha_policy",
    "load_flowmag_model",
    "unwrap",
]



# Copyright (c) Facebook, Inc. and its affiliates.
# https://github.com/facebookresearch/detectron2/blob/main/projects/TridentNet/tridentnet/trident_conv.py

import torch
from torch import nn
from torch.nn import functional as F
from torch.nn.modules.utils import _pair


class MultiScaleTridentConv(nn.Module):
    def __init__(
        self,
        in_channels,
        out_channels,
        kernel_size,
        stride=1,
        strides=1,
        paddings=0,
        dilations=1,
        dilation=1,
        groups=1,
        num_branch=1,
        test_branch_idx=-1,
        bias=False,
        norm=None,
        activation=None,
    ):
        super(MultiScaleTridentConv, self).__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.kernel_size = _pair(kernel_size)
        self.num_branch = num_branch
        self.stride = _pair(stride)
        self.groups = groups
        self.with_bias = bias
        self.dilation = dilation
        if isinstance(paddings, int):
            paddings = [paddings] * self.num_branch
        if isinstance(dilations, int):
            dilations = [dilations] * self.num_branch
        if isinstance(strides, int):
            strides = [strides] * self.num_branch
        self.paddings = [_pair(padding) for padding in paddings]
        self.dilations = [_pair(dilation) for dilation in dilations]
        self.strides = [_pair(stride) for stride in strides]
        self.test_branch_idx = test_branch_idx
        self.norm = norm
        self.activation = activation

        assert len({self.num_branch, len(self.paddings), len(self.strides)}) == 1

        self.weight = nn.Parameter(
            torch.Tensor(out_channels, in_channels // groups, *self.kernel_size)
        )
        if bias:
            self.bias = nn.Parameter(torch.Tensor(out_channels))
        else:
            self.bias = None

        nn.init.kaiming_uniform_(self.weight, nonlinearity="relu")
        if self.bias is not None:
            nn.init.constant_(self.bias, 0)

    def forward(self, inputs):
        num_branch = (
            self.num_branch if self.training or self.test_branch_idx == -1 else 1
        )
        assert len(inputs) == num_branch

        if self.training or self.test_branch_idx == -1:
            outputs = [
                F.conv2d(
                    input,
                    self.weight,
                    self.bias,
                    stride,
                    padding,
                    self.dilation,
                    self.groups,
                )
                for input, stride, padding in zip(inputs, self.strides, self.paddings)
            ]
        else:
            outputs = [
                F.conv2d(
                    inputs[0],
                    self.weight,
                    self.bias,
                    self.strides[self.test_branch_idx]
                    if self.test_branch_idx == -1
                    else self.strides[-1],
                    self.paddings[self.test_branch_idx]
                    if self.test_branch_idx == -1
                    else self.paddings[-1],
                    self.dilation,
                    self.groups,
                )
            ]

        if self.norm is not None:
            outputs = [self.norm(x) for x in outputs]
        if self.activation is not None:
            outputs = [self.activation(x) for x in outputs]
        return outputs



import numpy as np
import os
import pylab
import webbrowser
from PIL import Image, ImageDraw, ImageFont
from neurovc.contrib.flowmag.flow_models.pwcnet.aolib import util as ut

# import scipy.misc, scipy.misc.pilutil # not sure if this is necessary
import scipy.ndimage
from io import BytesIO
# from io import StringIO
# from io import BytesIO; StringIO = BytesIO
# import cv


def show(*args, **kwargs):
    from neurovc.contrib.flowmag.flow_models.pwcnet.aolib import imtable

    return imtable.show(*args, **kwargs)


# Functional code for drawing on images:
def draw_on(f, im):
    pil = to_pil(im)
    draw = ImageDraw.ImageDraw(pil)
    f(draw)
    return from_pil(pil)


def color_from_string(s):
    """todo: add more, see matplotlib.colors.cnames"""
    colors = {"r": (255, 0, 0), "g": (0, 255, 0), "b": (0, 0, 255)}
    if s in colors:
        return colors[s]
    else:
        ut.fail("unknown color: %s" % s)


def parse_color(c):
    if isinstance(c, tuple) or isinstance(c, np.ndarray):
        return c
    elif isinstance(c, str):
        return color_from_string(c)


def colors_from_input(color_input, default, n):
    """Parse color given as input argument; gives user several options"""
    # todo: generalize this to non-colors
    expanded = None
    if color_input is None:
        expanded = [default] * n
    elif (isinstance(color_input, tuple)) and list(map(type, color_input)) == [
        int,
        int,
        int,
    ]:
        # expand (r, g, b) -> [(r, g, b), (r, g, b), ..]
        expanded = [color_input] * n
    else:
        # general case: [(r1, g1, b1), (r2, g2, b2), ...]
        expanded = color_input

    expanded = list(map(parse_color, expanded))
    return expanded


def draw_rects(
    im,
    rects,
    outlines=None,
    fills=None,
    texts=None,
    text_colors=None,
    line_widths=None,
    as_oval=False,
):
    rects = list(rects)
    outlines = colors_from_input(outlines, (0, 0, 255), len(rects))
    text_colors = colors_from_input(text_colors, (255, 255, 255), len(rects))
    fills = colors_from_input(fills, None, len(rects))

    if texts is None:
        texts = [None] * len(rects)
    if line_widths is None:
        line_widths = [None] * len(rects)

    def check_size(x, s):
        ut.check(x is None or len(x) == len(rects), "%s different size from rects" % s)

    check_size(outlines, "outlines")
    check_size(fills, "fills")
    check_size(texts, "texts")
    check_size(text_colors, "texts")

    def f(draw):
        for (x, y, w, h), outline, fill, text, text_color, lw in zip(
            rects, outlines, fills, texts, text_colors, line_widths
        ):
            if lw is None:
                if as_oval:
                    draw.ellipse((x, y, x + w, y + h), outline=outline, fill=fill)
                else:
                    draw.rectangle((x, y, x + w, y + h), outline=outline, fill=fill)
            else:
                # TODO: to do this right, we need to find where PIL draws the corners
                # x -= lw
                # y -= lw
                # w += 2*lw
                # h += 2*lw
                # pts = [(x, y), (x + w, y), (x + w, y + h), (x, y + h)]
                # for i in xrange(len(pts)):
                #   #draw.line(pts[i] + pts[(i+1)%4], fill = outline, width = lw)
                #   draw.rectangle(pts[i] + pts[(i+1)%4], fill = outline, width = lw)
                d = int(np.ceil(lw / 2))
                draw.rectangle((x - d, y - d, x + w + d, y + d), fill=outline)
                draw.rectangle((x - d, y - d, x + d, y + h + d), fill=outline)

                draw.rectangle((x + w + d, y + h + d, x - d, y + h - d), fill=outline)
                draw.rectangle((x + w + d, y + h + d, x + w - d, y - d), fill=outline)

            if text is not None:
                # draw text inside rectangle outline
                border_width = 2
                draw.text((border_width + x, y), text, fill=text_color)

    return draw_on(f, im)


def draw_rects_scale(
    sc, im, rects, outlines=None, fills=None, texts=None, text_colors=None
):
    scaled_rects = []
    for r in rects:
        r = np.array(r)
        sr = r * sc
        if r[2] >= 1 and r[3] >= 1:
            sr[2:] = np.maximum(sr[2:], 1.0)
        scaled_rects.append(sr)

    return draw_rects(scale(im, sc), scaled_rects, outlines, fills, texts, text_colors)


def draw_pts(im, points, colors=None, width=1, texts=None):
    # ut.check(colors is None or len(colors) == len(points))
    points = list(points)
    colors = colors_from_input(colors, (255, 0, 0), len(points))
    rects = [(p[0] - width / 2, p[1] - width / 2, width, width) for p in points]
    return draw_rects(
        im, rects, fills=colors, outlines=[None] * len(points), texts=texts
    )


def draw_lines(im, pts1, pts2, colors=None, width=0):
    ut.check(len(pts1) == len(pts2), "Line endpoints different sizes")
    colors = colors_from_input(colors, None, len(pts1))

    def f(draw):
        for p1, p2, c in zip(pts1, pts2, colors):
            draw.line(ut.int_tuple(p1) + ut.int_tuple(p2), fill=c, width=width)

    return draw_on(f, im)


def draw_text(im, texts, pts, colors, font_size=None, bold=False):
    im = rgb_from_gray(im)
    # todo: add fonts, call from draw_rects
    ut.check(len(pts) == len(texts))
    # ut.check((colors is None) or len(colors) == len(texts))
    colors = colors_from_input(colors, (0, 0, 0), len(texts))

    def f(draw):
        if font_size is None:
            font = None
        else:
            # font_name = '/usr/share/fonts/truetype/ttf-liberation/LiberationMono-Regular.ttf'
            font_choices = [
                "/usr/share/fonts/truetype/freefont/FreeMono%s.ttf"
                % ("Bold" if bold else ""),
                "/Library/Fonts/PTMono.ttc",
            ]
            for font_name in font_choices:
                if os.path.exists(font_name):
                    break
            else:
                raise RuntimeError(
                    "could not find a suitable font on this machine (please edit paths in img.py)"
                )

            font = ImageFont.truetype(font_name, size=font_size)

        for pt, text, color in zip(pts, texts, colors):
            draw.text(ut.int_tuple(pt), text, fill=color, font=font)

    return draw_on(f, im)


def draw_text_ul(im, text, color=(0, 255, 0), font_size=25):
    return draw_text(im, [text], [(0, 0)], [color], font_size=font_size)


def luminance(im):
    if len(im.shape) == 2:
        return im
    else:
        # see http://www.mathworks.com/help/toolbox/images/ref/rgb2gray.html
        return np.uint8(
            np.round(0.2989 * im[:, :, 0] + 0.587 * im[:, :, 1] + 0.114 * im[:, :, 2])
        )


# def sub_img(im, x_or_rect, y = None, w = None, h = None):
def sub_img(im, x_or_rect, y=None, w=None, h=None):
    if x_or_rect is None:
        return im
    elif y is None:
        x, y, w, h = x_or_rect
    else:
        x = x_or_rect
    return im[y : y + h, x : x + w]


def sub_img_frac(im, x_or_rect, y=None, w=None, h=None):
    if y is None:
        x, y, w, h = x_or_rect
    else:
        x = x_or_rect
    x = int(x * im.shape[1])
    y = int(y * im.shape[0])
    w = int(w * im.shape[1])
    h = int(h * im.shape[0])
    return im[y : y + h, x : x + w]


# def stack_img_pair(im1, im2):
#   h1, w1 = im1.shape[:2]
#   h2, w2 = im2.shape[:2]
#   im3 = np.zeros((max(h1, h2), w1 + w2, 3), dtype = im1.dtype)
#   im3[:h1, :w1, :] = rgb_from_gray(im1)
#   im3[:h2, w1:, :] = rgb_from_gray(im2)
#   return im3

# def stack_imgs(ims):
#   """ slow, should rewrite """
#   assert len(ims) > 0
#   res = ims[0]
#   for im in ims[1:]:
#     res = stack_img_pair(res, im)
#   return res

# def hstack_ims(ims):
#   max_h = max(im.shape[0] for im in ims)
#   result = []
#   for im in ims:
#     frame = np.zeros((max_h, im.shape[1], 3))
#     frame[:im.shape[0],:im.shape[1]] = rgb_from_gray(im)
#     result.append(frame)
#   return np.hstack(result)


def hstack_ims(ims, bg_color=(0, 0, 0)):
    max_h = max([im.shape[0] for im in ims])
    result = []
    for im in ims:
        # frame = np.zeros((max_h, im.shape[1], 3))
        frame = make(im.shape[1], max_h, bg_color)
        frame[: im.shape[0], : im.shape[1]] = rgb_from_gray(im)
        result.append(frame)
    return np.hstack(result)


# def hstack_ims_mult(*all_ims):
#   max_h = max(max(im.shape[0] for im in ims) for ims in all_ims)

#   result = []
#   for im in ims:
#     frame = np.zeros((max_h, im.shape[1], 3))
#     frame[:im.shape[0],:im.shape[1]] = rgb_from_gray(im)
#     result.append(frame)
#   return np.hstack(result)


def vstack_ims(ims, bg_color=(0, 0, 0)):
    if len(ims) == 0:
        return make(0, 0)

    max_w = max([im.shape[1] for im in ims])
    result = []
    for im in ims:
        # frame = np.zeros((im.shape[0], max_w, 3))
        frame = make(max_w, im.shape[0], bg_color)
        frame[: im.shape[0], : im.shape[1]] = rgb_from_gray(im)
        result.append(frame)
    return np.vstack(result)


def make_rgb(im):
    im = rgb_from_gray(im, False)
    if im.shape[2] < 3:
        raise RuntimeError()
    elif im.shape[2] > 3:
        im = im[:, :, :3]
    return im


def rgb_from_gray(img, copy=True, remove_alpha=True):
    if img.ndim == 3 and img.shape[2] == 3:
        return img.copy() if copy else img
    elif img.ndim == 3 and img.shape[2] == 4:
        return (img.copy() if copy else img)[..., :3]
    elif img.ndim == 3 and img.shape[2] == 1:
        return np.tile(img, (1, 1, 3))
    elif img.ndim == 2:
        return np.tile(img[:, :, np.newaxis], (1, 1, 3))
    else:
        raise RuntimeError("Cannot convert to rgb. Shape: " + str(img.shape))


def load(im_fname, gray=False):
    if im_fname.endswith(".gif"):
        print("GIFs don't load correctly for some reason")
        ut.fail("fail")
    im = from_pil(Image.open(im_fname))
    # use imread, then flip upside down
    # im = np.array(list(reversed(pylab.imread(im_fname)[:,:,:3])))
    if gray:
        return luminance(im)
    elif not gray and np.ndim(im) == 2:
        return rgb_from_gray(im)
    else:
        return im


imread = load


def loadsc(fname, scale, gray=False):
    return resize(load(fname, gray=gray), scale)


def save(img_fname, a):
    if img_fname.endswith("jpg"):
        return Image.fromarray(np.uint8(a)).save(img_fname, quality=100)
    else:
        # return Image.fromarray(np.uint8(a)).save(img_fname)
        return Image.fromarray(np.uint8(a)).save(img_fname, quality=100)


# def make_temp_file(ext):
#   fd, fname = tempfile.mkstemp(ext)
#   # shouldn't delete file
#   os.close(fd)
#   return fname

# def make_pretty(img):
#   if img.dtype == 'bool':
#     return img * 255
#   elif (0 <= np.min(img)) and (np.max(img) <= 1.0):
#     return img*255
#   return img


def show_html(html):
    page = ut.make_temp(".html")
    ut.make_file(page, html)
    print("opening", page)
    webbrowser.open(page)


# # http://opencv.willowgarage.com/wiki/PythonInterface
# def cv2array(im):
#   depth2dtype = {
#       cv.IPL_DEPTH_8U: 'uint8',
#       cv.IPL_DEPTH_8S: 'int8',
#       cv.IPL_DEPTH_16U: 'uint16',
#       cv.IPL_DEPTH_16S: 'int16',
#       cv.IPL_DEPTH_32S: 'int32',
#       cv.IPL_DEPTH_32F: 'float32',
#       cv.IPL_DEPTH_64F: 'float64',
#     }

#   arrdtype=im.depth
#   a = np.fromstring(
#        im.tostring(),
#        dtype=depth2dtype[im.depth],
#        count=im.width*im.height*im.nChannels)
#   a.shape = (im.height,im.width,im.nChannels)
#   return a

# def to_cv(a):
#   dtype2depth = {
#       'uint8':   cv.IPL_DEPTH_8U,
#       'int8':  cv.IPL_DEPTH_8S,
#       'uint16':  cv.IPL_DEPTH_16U,
#       'int16':   cv.IPL_DEPTH_16S,
#       'int32':   cv.IPL_DEPTH_32S,
#       'float32': cv.IPL_DEPTH_32F,
#       'float64': cv.IPL_DEPTH_64F,
#     }
#   try:
#     nChannels = a.shape[2]
#   except:
#     nChannels = 1
#   cv_im = cv.CreateImageHeader((a.shape[1],a.shape[0]),
#       dtype2depth[str(a.dtype)],
#       nChannels)
#   cv.SetData(cv_im, a.tostring(),
#          a.dtype.itemsize*nChannels*a.shape[1])
#   return cv_im


# def to_pil(im): return Image.fromarray(np.uint8(im))
def to_pil(im):
    # print im.dtype
    return Image.fromarray(np.uint8(im))


def from_pil(pil):
    # print pil
    return np.array(pil)


def to_pylab(a):
    return np.uint8(a)


def test_draw_text():
    im = 255 + np.zeros((300, 300, 3))
    show(
        [
            draw_text(
                im, ["hello", "world"], [(100, 200), (0, 0)], [(255, 0, 0), (0, 255, 0)]
            ),
            draw_text(
                im,
                ["hello", "world"],
                [(100, 100), (0, 0)],
                [(255, 0, 0), (0, 255, 0)],
                font_size=12,
            ),
        ]
    )


def save_tmp(im, encoding=".png", dir=None):
    fname = ut.make_temp(encoding, dir=dir)
    save(fname, im)
    return fname


def save_tmp_nfs(im, encoding=".png"):
    return save_tmp(im, encoding, "/csail/vision-billf5/aho/tmp")


# def resize(im, size):
#   if type(size) == type(1):
#     size = float(size)
#   #return scipy.misc.pilutil.imresize(im, size)
#   return scipy.misc.imresize(im, size)


# def resize(im, scale, order = 3, hires = 'auto'):


def resize(im, scale, order=3, hires=False):
    if hires == "auto":
        hires = im.dtype == np.uint8

    if np.ndim(scale) == 0:
        new_scale = [scale, scale]
    # interpret scale as dimensions; convert integer size to a fractional scale
    elif (
        ((scale[0] is None) or isinstance(scale[0], int))
        and ((scale[1] is None) or isinstance(scale[1], int))
        and (not (scale[0] is None and scale[1] is None))
    ):
        # if the size of only one dimension is provided, scale the other to maintain the right aspect ratio
        if scale[0] is None:
            dims = (int(float(im.shape[0]) / im.shape[1] * scale[1]), scale[1])
        elif scale[1] is None:
            dims = (scale[0], int(float(im.shape[1]) / im.shape[0] * scale[0]))
        else:
            dims = scale[:2]

        new_scale = [
            float(dims[0] + 0.4) / im.shape[0],
            float(dims[1] + 0.4) / im.shape[1],
        ]
        # a test to make sure we set the floating point scale correctly
        result_dims = [int(new_scale[0] * im.shape[0]), int(new_scale[1] * im.shape[1])]
        assert tuple(result_dims) == tuple(dims)
    elif isinstance(scale[0], float) and isinstance(scale[1], float):
        new_scale = scale
        # new_scale = scale[1], scale[0]
    else:
        raise RuntimeError("don't know how to interpret scale: %s" % (scale,))
        # want new scale' to be such that
        # int(scale'[0]*im.shape[0]) = scale[0], etc. (that's how zoom computes the new shape)
        # todo: any more numerical issues?
        # print 'scale before', im.shape, scale
        # print 'scale after', scale
        # print 'new image size', [int(scale[0]*im.shape[0]),int(scale[1]*im.shape[1])]
    # scale_param = new_scale if im.ndim == 2 else (new_scale[0], new_scale[1], 1)
    scale_param = new_scale if im.ndim == 2 else (new_scale[0], new_scale[1], 1)

    if hires:
        # sz = map(int, (scale_param*im.shape[1], scale_param*im.shape[0]))
        sz = list(
            map(int, (scale_param[1] * im.shape[1], scale_param[0] * im.shape[0]))
        )
        return from_pil(to_pil(im).resize(sz, Image.ANTIALIAS))
    else:
        res = scipy.ndimage.zoom(im, scale_param, order=order)
        # verify that zoom() returned an image of the desired size
        if (
            (np.ndim(scale) != 0)
            and isinstance(scale[0], int)
            and isinstance(scale[1], int)
        ):
            assert res.shape[:2] == (scale[0], scale[1])
        return res


scale = resize

# import skimage
# resize = skimage.imresize


def test_resize():
    im = make(44, 44)
    assert resize(im, (121, 120, "t")).shape[:2] == (121, 120)
    assert resize(im, (2.0, 0.5, "t")).shape[:2] == (88, 22)


def show_file(fname):
    show(load(fname))


def img_extensions():
    return ["png", "gif", "jpg", "jpeg", "bmp", "ppm", "pgm"]


def is_img_file(fname):
    return any(fname.lower().endswith(ext) for ext in img_extensions())


def blur(im, sigma):
    if np.ndim(im) == 2:
        return scipy.ndimage.filters.gaussian_filter(im, sigma)
    else:
        return np.concatenate(
            [
                scipy.ndimage.filters.gaussian_filter(im[:, :, i], sigma)[
                    :, :, np.newaxis
                ]
                for i in range(im.shape[2])
            ],
            axis=2,
        )


def blit(src, dst, x, y, opt=None):
    if opt == "center":
        x -= src.shape[1] / 2
        y -= src.shape[0] / 2
    # crop intersecting
    dx, dy, dw, dh = ut.crop_rect_to_img((x, y, src.shape[1], src.shape[0]), dst)
    sx = dx - x
    sy = dy - y

    dst[dy : dy + dh, dx : dx + dw] = src[sy : sy + dh, sx : sx + dw]


def weighted_add(src, dst, x, y, src_weight, dst_weight, opt=None):
    if opt == "center":
        x -= src.shape[1] / 2
        y -= src.shape[0] / 2
    # crop intersecting
    dx, dy, dw, dh = ut.crop_rect_to_img((x, y, src.shape[1], src.shape[0]), dst)
    sx = dx - x
    sy = dy - y
    dst[dy : dy + dh, dx : dx + dw] = (
        dst[dy : dy + dh, dx : dx + dw] * dst_weight
        + src[sy : sy + dh, sx : sx + dw] * src_weight
    )


def make(w, h, fill=(0, 0, 0)):
    return np.uint8(np.tile([[fill]], (h, w, 1)))


def luminance_rgb(im):
    return rgb_from_gray(luminance(im))


def rotate(img, angle, fill=0):
    """Rotate image around its center by the given angle (in
    radians).  No interpolation is used; indices are rounded.  The
    returned image may be larger than the original, but the middle
    pixel corresponds to the middle of the original.  Pixels with no
    correspondence are filled as 'fill'.

    Also returns mapping from original image to rotated."""
    r = int(np.ceil(np.sqrt(img.shape[0] ** 2 + img.shape[1] ** 2)))
    X, Y = np.mgrid[0:r, 0:r]
    X = X.flatten()
    Y = Y.flatten()
    X2 = np.array(
        np.round(
            img.shape[1] / 2 + np.cos(angle) * (X - r / 2) - np.sin(angle) * (Y - r / 2)
        ),
        dtype=int,
    )
    Y2 = np.array(
        np.round(
            img.shape[0] / 2 + np.sin(angle) * (X - r / 2) + np.cos(angle) * (Y - r / 2)
        ),
        dtype=int,
    )
    good = ut.logical_and_many(X2 >= 0, X2 < img.shape[1], Y2 >= 0, Y2 < img.shape[0])
    out = fill + np.zeros(
        (r, r) if img.ndim == 2 else (r, r, img.shape[2]), dtype=img.dtype
    )
    out[Y[good], X[good]] = img[Y2[good], X2[good]]
    T = np.dot(
        np.dot(
            ut.rigid_transform(np.eye(2), [img.shape[1] / 2, img.shape[0] / 2]),
            ut.rigid_transform(ut.rotation_matrix2(angle)),
        ),
        ut.rigid_transform(np.eye(2), [-r / 2, -r / 2]),
    )
    return out, np.linalg.inv(T)


def map_img(f, im, dtype=None, components=None):
    new_im = np.zeros(
        im.shape if components is None else im.shape + (components,),
        dtype=im.dtype if dtype is None else dtype,
    )
    for y in range(im.shape[0]):
        for x in range(im.shape[1]):
            new_im[y, x] = f(im[y, x])
    return new_im


def add_border(img, w, h, color=(0, 0, 0)):
    assert 0 <= w
    assert 0 <= h
    out = make(img.shape[1] + 2 * w, img.shape[0] + 2 * h, color)
    out[h : (h + img.shape[0]), w : (w + img.shape[1])] = img
    return out


def pad_corner(im, pw, ph, color=(0, 0, 0)):
    out = make(im.shape[1] + pw, im.shape[0] + ph, color)
    out[: im.shape[0], : im.shape[1]] = im
    return out


def expand(im, new_shape, opt="center"):
    if isinstance(new_shape, float):
        new_w = int(im.shape[1] * new_shape)
        new_h = int(im.shape[0] * new_shape)
    elif isinstance(new_shape, tuple):
        new_shape = new_shape[:2]
        new_h, new_w = new_shape
    else:
        raise RuntimeError("Don't know how to interpret shape")

    if im.shape[0] >= new_h and im.shape[1] >= new_w:
        return im.copy()
    else:
        im = rgb_from_gray(im)
        r = make(new_w, new_h)
        if opt == "center":
            blit(im, r, im.shape[1] / 2, im.shape[0] / 2, opt="center")
        elif opt == "corner":
            r[: im.shape[0], : im.shape[1]] = im
        return r


def combine_rgb(r, g, b):
    a = np.zeros(r.shape + (3,))
    a[:, :, 0] = r
    a[:, :, 1] = g
    a[:, :, 2] = b
    return a


def compute_pyramid(ptm, interval, min_size):
    # based on pff's featpyramid.m
    # todo: upsample one level
    sc = 2 ** (1.0 / interval)
    imsize = ptm.shape[:2]
    max_scale = int(1 + np.floor(np.log(np.min(imsize) / min_size) / np.log(sc)))
    ims = [None] * max_scale
    scale = [None] * len(ims)

    # skipping 2x scale
    for i in range(1, interval + 1):
        im_scaled = resize(ptm, 1 / sc ** (i - 1))
        ims[-1 + i] = im_scaled
        scale[-1 + i] = 1 / sc ** (i - 1)
        for j in range(i + interval, max_scale + 1, interval):
            im_scaled = resize(im_scaled, 0.5)
            ims[-1 + j] = im_scaled
            scale[-1 + j] = 0.5 * scale[-1 + j - interval]
    assert None not in ims
    return ims, scale


# imrotate = scipy.misc.imrotate


def imrotate(*args):
    import warnings

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        return scipy.misc.imrotate(*args)


def from_fig_slow(fig=None, tight=True):
    ext = "png"
    if fig is None:
        fig = pylab.gcf()
    IO = BytesIO()
    if tight:
        pylab.savefig(IO, format=ext, bbox_inches="tight")
    else:
        pylab.savefig(IO, format=ext)
    IO.seek(0)
    return from_pil(Image.open(IO))


# def from_fig_fast(fig = None, tight = True):
#   ext = 'raw'
#   if fig is None:
#     fig = pylab.gcf()
#   IO = BytesIO()
#   pylab.savefig(IO, format = ext)
#   IO.seek(0)
#   w, h = fig.canvas.get_width_height()
#   return np.fromstring(IO.buf, dtype = np.uint8).reshape((600, -1, 4))

# def from_fig(fig = None):
#   """
#   http://www.icare.univ-lille1.fr/wiki/index.php/How_to_convert_a_matplotlib_figure_to_a_numpy_array_or_a_PIL_image
#   @brief Convert a Matplotlib figure to a 4D numpy array with RGBA channels and return it
#   @param fig a matplotlib figure
#   @return a numpy 3D array of RGBA values
#   """

#   if fig is None:
#     fig = pylab.gcf()

#   # draw the renderer
#   fig.canvas.draw()

#   # Get the RGBA buffer from the figure
#   w,h = fig.canvas.get_width_height()
#   buf = np.fromstring(fig.canvas.tostring_argb(), dtype=np.uint8)
#   buf.shape = (h, w, 4)

#   # canvas.tostring_argb give pixmap in ARGB mode. Roll the ALPHA channel to have it in RGBA mode
#   buf = np.roll(buf, 3, axis = 2)
#   return buf
#   #return buf[..., 1:]

# def from_fig(fig = None):
#   """
#   @brief Convert a Matplotlib figure to a 4D numpy array with RGBA channels and return it
#   @param fig a matplotlib figure
#   @return a numpy 3D array of RGBA values
#   http://www.icare.univ-lille1.fr/wiki/index.php/How_to_convert_a_matplotlib_figure_to_a_numpy_array_or_a_PIL_image
#   """
#   if fig is None:
#     fig = pylab.gcf()

#   # draw the renderer
#   fig.canvas.draw()

#   # Get the RGBA buffer from the figure
#   w,h = fig.canvas.get_width_height()
#   buf = np.fromstring(fig.canvas.tostring_argb(), dtype=np.uint8)
#   buf.shape = (h, w, 4)

#   # canvas.tostring_argb give pixmap in ARGB mode. Roll the ALPHA channel to have it in RGBA mode
#   buf = np.roll(buf, 3, axis = 2)

#   # not sure how to set the background to white
#   p =  buf[:, :, 3] / 255.
#   buf = np.array(buf[:, :, :3] * p[:, :, np.newaxis] + (1 - p)[:, :, np.newaxis]*255, 'uint8')
#   return buf


def from_fig(fig=None, size_inches=None):
    """
    @brief Convert a Matplotlib figure to a 4D numpy array with RGBA channels and return it
    @param fig a matplotlib figure
    @return a numpy 3D array of RGBA values
    http://www.icare.univ-lille1.fr/wiki/index.php/How_to_convert_a_matplotlib_figure_to_a_numpy_array_or_a_PIL_image
    """
    if fig is None:
        fig = pylab.gcf()
    if size_inches is not None:
        fig.set_size_inches(*size_inches)
    # draw the renderer
    fig.canvas.draw()

    # Get the RGBA buffer from the figure
    w, h = fig.canvas.get_width_height()
    buf = np.fromstring(fig.canvas.tostring_argb(), dtype=np.uint8)
    buf.shape = (h, w, 4)

    # canvas.tostring_argb give pixmap in ARGB mode. Roll the ALPHA channel to have it in RGBA mode
    buf = np.roll(buf, 3, axis=2)

    # not sure how to set the background to white
    p = buf[:, :, 3] / 255.0
    buf = np.array(
        buf[:, :, :3] * p[:, :, np.newaxis] + (1 - p)[:, :, np.newaxis] * 255, "uint8"
    )
    return buf


def show_fig():
    show(from_fig())


def scale_vals(A, lo, hi):
    return np.uint8(255 * (np.clip(A, lo, hi) - lo) / float(hi - lo))


def merge_ims(srcs, pts_or_rects, bg, opt=None):
    """Makes a new image where each image in patches is copied at a
    corresponding pixel location.  Overlapping images are averaged
    together."""
    dst = rgb_from_gray(bg)
    layer = np.zeros(dst.shape)
    # counts = np.zeros(dst.shape[:2], 'l')
    counts = np.zeros(dst.shape[:2], "d")
    for src, r in zip(srcs, pts_or_rects):
        r = ut.int_tuple(r)
        x, y = r[:2]

        # rescale if we're given a rectangle, and it has a different size
        if len(r) > 2:
            assert len(r) == 4
            assert opt != "center"
            if src.shape[:2] != (r[3], r[2]):
                src = resize(src, (r[3], r[2]))
        elif opt == "center":
            x -= src.shape[1] / 2
            y -= src.shape[0] / 2

        # crop intersecting
        dx, dy, dw, dh = ut.crop_rect_to_img((x, y, src.shape[1], src.shape[0]), dst)
        sx = dx - x
        sy = dy - y
        layer[dy : dy + dh, dx : dx + dw] += src[sy : sy + dh, sx : sx + dw, :3]
        if np.ndim(src) == 3 and src.shape[2] == 4:
            counts[dy : dy + dh, dx : dx + dw] += (
                np.array(src[sy : sy + dh, sx : sx + dw, 3], "d") / 255.0
            )
        else:
            counts[dy : dy + dh, dx : dx + dw] += 1
    dst[counts > 0] = layer[counts > 0] / counts[counts > 0][:, np.newaxis]
    return dst


def label_im(im, text, color=(0, 255, 0)):
    return draw_text(im, [text], [(25, im.shape[0] - 25)], [color])


def remap_color(im, xy):
    assert im.shape[:2] == xy.shape[:2]
    assert xy.shape[2] == 2
    vals = []
    for i in range(im.shape[2]):
        dx = xy[..., 0].flatten()[np.newaxis, :]
        dy = xy[..., 1].flatten()[np.newaxis, :]
        v = scipy.ndimage.map_coordinates(im[..., i], np.concatenate([dy, dx]))
        vals.append(v.reshape(im.shape[:2] + (1,)))
    return np.concatenate(vals, axis=2)


def stack_meshgrid(xs, ys, dtype="l"):
    x, y = np.meshgrid(xs, ys)
    return np.array(
        np.concatenate([x[..., np.newaxis], y[..., np.newaxis]], axis=2), dtype=dtype
    )


def sub_img_pad(im, xxx_todo_changeme, oob=0):
    (x, y, w, h) = xxx_todo_changeme
    if len(im.shape) == 2:
        dst = np.zeros((h, w))
    else:
        dst = np.zeros((h, w, im.shape[2]))

    dst[:] = oob
    sx, sy, sw, sh = ut.crop_rect_to_img((x, y, w, h), im)
    dst[(sy - y) : (sy - y) + sh, (sx - x) : (sx - x) + sw] = im[
        sy : sy + sh, sx : sx + sw
    ]
    return dst


def sub_img_reflect(im, xxx_todo_changeme1):
    (x, y, w, h) = xxx_todo_changeme1
    x, y, w, h = list(map(ut.iround, [x, y, w, h]))
    yy, xx = np.mgrid[y : y + h, x : x + w]
    vals = np.uint8(
        lookup_bilinear(im, xx.flatten(), yy.flatten(), order=0, mode="reflect")
    )
    return vals.reshape((h, w, im.shape[2]))


def compress(im, format="png"):
    out = BytesIO()
    im = to_pil(im)
    im.save(out, format=format)
    c = out.getvalue()
    out.close()
    return c


def compress_jpeg(im, format="jpeg"):
    return compress(im, format)


# def uncompress(s):
#   return from_pil(Image.open(BytesIO(s)))


def uncompress(s):
    # return from_pil(Image.open(BytesIO(s.encode())))
    if isinstance(s, str):
        return from_pil(Image.open(BytesIO(s.encode("latin-1"))))
    else:
        return from_pil(Image.open(BytesIO(s)))


# def cv_uncompress(s):
#   import cv2
#   #return cv2.imdecode(s)

#   a = cv2.imdecode(np.fromstring(s, np.uint8), cv2.IMREAD_COLOR)#
#   b = uncompress(s)
#   print a.shape, b.shape
#   print np.mean(np.abs(a.astype('float32')-b.astype('float32')))
#   return a


def test_compress():
    im = load("/afs/csail.mit.edu/u/a/aho/bear.jpg")
    print("orig", ut.guess_bytes(im))
    s = compress(im)
    print("comp", ut.guess_bytes(s))
    assert np.all(im == uncompress(s))


def mix_ims(im1, im2, mask, alpha=0.5):
    im1 = im1.copy()
    im2 = np.asarray(im2)
    if len(im2) == 3:
        # single color
        im1[mask] = im1[mask] * alpha + im2 * (1 - alpha)
    else:
        im1[mask] = im1[mask] * alpha + im2[mask] * (1 - alpha)
    return im1


# def lookup_bilinear(im, x, y, order = 3, mode = 'constant', cval = 0.0):
def lookup_bilinear(im, x, y, order=1, mode="constant", cval=0.0):
    yx = np.array([y, x])
    if np.ndim(im) == 2:
        return scipy.ndimage.map_coordinates(im, yx, order=order, mode=mode, cval=cval)
    else:
        return np.concatenate(
            [
                scipy.ndimage.map_coordinates(im[:, :, i], yx, order=order, mode=mode)[
                    :, np.newaxis
                ]
                for i in range(im.shape[2])
            ],
            axis=1,
        )


def map_helper(xxx_todo_changeme2):
    (xs, i, order, mode) = xxx_todo_changeme2
    return scipy.ndimage.map_coordinates(xs, np.array([i]), order=order, mode=mode)[
        :, np.newaxis
    ]


def lookup_bilinear1d(xs, i, order=4, mode="constant", cval=0.0, par=0):
    if np.ndim(xs) == 1:
        return scipy.ndimage.map_coordinates(xs, i, order=order, mode=mode, cval=cval)
    else:
        if par:
            vals = [(xs[:, j], i, order, mode) for j in range(xs.shape[1])]
            return np.concatenate(ut.parmap(map_helper, vals), axis=1)
        else:
            return np.concatenate(
                [
                    scipy.ndimage.map_coordinates(
                        xs[:, j], np.array([i]), order=order, mode=mode
                    )[:, np.newaxis]
                    for j in range(xs.shape[1])
                ],
                axis=1,
            )


# def pixels_in_bounds(im, xs, ys):
def pixels_in_bounds(im_shape, xs, ys):
    return ut.land(0 <= xs, xs < im_shape[1], 0 <= ys, ys < im_shape[0])


def im2float(im):
    im = np.array(im, "float32")
    im /= 255.0
    return im


def try_load_img(fname, default_size=(256, 256)):
    try:
        return load(fname)
    except IOError:
        print("Failed to load:", fname)
        return make(default_size[0], default_size[1])


def exif(fname, use_jhead=True):
    if use_jhead:
        pass
    else:
        with Image.open(fname) as img:
            if not hasattr(img, "_getexif"):
                return None
            return img._getexif()


def exif_shape(fname):
    ex = exif(fname)
    yid = 40963
    xid = 40962
    if ex is None or xid not in ex or yid not in ex:
        return None
    return ex[yid], ex[xid]



import torch
import torch.nn as nn
import torch.nn.functional as F


class Correlation(nn.Module):
    def __init__(self, max_displacement=4, *args, **kwargs):
        super(Correlation, self).__init__()
        self.max_displacement = max_displacement
        self.output_dim = 2 * self.max_displacement + 1
        self.pad_size = self.max_displacement

    def forward(self, x1, x2):
        B, C, H, W = x1.size()

        x2 = F.pad(x2, [self.pad_size] * 4)
        cv = []
        for i in range(self.output_dim):
            for j in range(self.output_dim):
                cost = x1 * x2[:, :, i : (i + H), j : (j + W)]
                cost = torch.mean(cost, 1, keepdim=True)
                cv.append(cost)
        return torch.cat(cv, 1)


if __name__ == "__main__":
    import time
    import random
    from correlation_package.correlation import Correlation as Correlation_cuda

    device = torch.device("cuda") if torch.cuda.is_available() else torch.device("cpu")
    corr1 = Correlation(
        max_displacement=4, kernel_size=1, stride1=1, stride2=1, corr_multiply=1
    ).to(device)

    corr2 = Correlation_cuda(
        pad_size=4,
        kernel_size=1,
        max_displacement=4,
        stride1=1,
        stride2=1,
        corr_multiply=1,
    )

    t1_sum = 0
    t2_sum = 0

    for i in range(50):
        C = random.choice([128, 256])
        H = random.choice([128, 256])  # , 512
        W = random.choice([64, 128])  # , 256
        x1 = torch.randn(4, C, H, W, requires_grad=True).to(device)
        x2 = torch.randn(4, C, H, W).to(device)

        end = time.time()
        y2 = corr2(x1, x2)
        t2_f = time.time() - end

        end = time.time()
        y2.sum().backward()
        t2_b = time.time() - end

        end = time.time()
        y1 = corr1(x1, x2)
        t1_f = time.time() - end

        end = time.time()
        y1.sum().backward()
        t1_b = time.time() - end

        assert torch.allclose(y1, y2, atol=1e-7)

        print(
            "Forward: cuda: {:.3f}ms, pytorch: {:.3f}ms".format(t1_f * 100, t2_f * 100)
        )
        print(
            "Backward: cuda: {:.3f}ms, pytorch: {:.3f}ms".format(t1_b * 100, t2_b * 100)
        )

        if i < 3:
            continue
        t1_sum += t1_b + t1_f
        t2_sum += t2_b + t2_f

    print("cuda: {:.3f}s, pytorch: {:.3f}s".format(t1_sum, t2_sum))
    ...



"""Streaming-friendly FlowMag wrappers compatible with momag-style interfaces."""

from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Callable, Dict, Iterable, Optional, Protocol, Tuple

import numpy as np
import torch
import torch.nn as nn
from omegaconf import OmegaConf
from pathlib import Path

from neurovc.contrib.flowmag.model import MotionMagModel


class OnlineMagnifier(Protocol):
    """Protocol matching OnlineMotionMagnifier ergonomics."""

    def update_reference(self, ref: Any, *, landmarks: Any | None = None) -> None: ...

    def __call__(
        self,
        frame: Any,
        *,
        alpha: float | None = None,
        mask: Any | None = None,
        landmarks: Any | None = None,
        depth: Any | None = None,
        return_info: bool = False,
    ) -> Any: ...


def unwrap(model: nn.Module) -> nn.Module:
    """Return the underlying module if wrapped by DP/DDP."""
    return model.module if hasattr(model, "module") else model


def default_alpha_policy(alpha: float, max_alpha: float) -> Tuple[float, int]:
    """Match FlowMag inference recursion (sqrt if above max_alpha)."""
    if alpha <= max_alpha:
        return alpha, 1
    a = alpha**0.5
    if a <= max_alpha:
        return a, 2
    raise ValueError(f"alpha={alpha} exceeds supported range (max={max_alpha})")


def _to_tensor(img: Any, device: torch.device) -> Tuple[torch.Tensor, bool]:
    """Convert HWC BGR/RGB/CHW to CHW float tensor; returns (tensor, swapped_to_rgb)."""
    swapped = False
    if isinstance(img, torch.Tensor):
        t = img
        if t.ndim == 3 and t.shape[0] in (1, 3):
            pass
        elif t.ndim == 3 and t.shape[-1] in (1, 3):
            t = t.permute(2, 0, 1)
    else:
        arr = np.asarray(img)
        if arr.ndim != 3 or arr.shape[-1] not in (1, 3):
            raise ValueError("expected HxWxC image with 1 or 3 channels")
        # assume OpenCV BGR input; swap to RGB for the model
        arr = arr[..., ::-1]
        swapped = True
        t = torch.from_numpy(arr).permute(2, 0, 1)
    t = t.float()
    if t.max() > 1.5:
        t = t / 255.0
    return t.to(device), swapped


def _to_numpy(img: torch.Tensor, ref_like: Any, *, swap_to_bgr: bool) -> np.ndarray:
    arr = img.detach().cpu().permute(1, 2, 0).numpy()
    if swap_to_bgr:
        arr = arr[..., ::-1]
    if isinstance(ref_like, np.ndarray) and ref_like.dtype == np.uint8:
        arr = np.clip(arr * 255.0, 0, 255).astype(np.uint8)
    return arr


def _crop_to_multiple(
    tensor: torch.Tensor, multiple: int = 8, target_hw: Tuple[int, int] | None = None
) -> torch.Tensor:
    _, h, w = tensor.shape
    if target_hw is None:
        h_new = (h // multiple) * multiple
        w_new = (w // multiple) * multiple
    else:
        h_new = min(h, target_hw[0])
        w_new = min(w, target_hw[1])
        h_new = (h_new // multiple) * multiple
        w_new = (w_new // multiple) * multiple
    return tensor[:, :h_new, :w_new]


def load_flowmag_model(
    *,
    config_path: str | None = None,
    checkpoint_path: str,
    device: str | torch.device = "cuda",
    training: bool = False,
) -> Tuple[nn.Module, float, Optional[int]]:
    """Build and load a FlowMag model; returns (model, max_alpha, epoch)."""
    if config_path is None or not Path(config_path).exists():
        # fall back to packaged default config
        import neurovc.contrib.flowmag as flowmag_pkg

        default_cfg = (
            Path(flowmag_pkg.__file__).resolve().parent
            / "configs"
            / "alpha16.color10.yaml"
        )
        if not default_cfg.exists():
            raise FileNotFoundError(
                f"Could not find FlowMag config at '{config_path}' "
                f"or default '{default_cfg}'"
            )
        config_path = str(default_cfg)

    config = OmegaConf.load(str(config_path))
    config.config = str(config_path)
    config.train.ngpus = 1
    config.train.is_training = bool(training)
    config.data.batch_size = 1

    dev = torch.device(device)
    model = MotionMagModel(config).to(dev)
    model = nn.DataParallel(model, device_ids=[0])
    chkpt = torch.load(checkpoint_path, map_location=dev)
    model.load_state_dict(chkpt["state_dict"], strict=False)
    max_alpha = float(config.train.alpha_high)
    epoch = chkpt.get("epoch")
    return model, max_alpha, epoch


class FlowMagOnline:
    """Stateful wrapper for frame-pair inference with FlowMag."""

    def __init__(
        self,
        model: nn.Module,
        *,
        device: str | torch.device = "cuda",
        max_alpha: float | None = None,
        default_alpha: float = 2.0,
        alpha_policy: Callable[
            [float, float], Tuple[float, int]
        ] = default_alpha_policy,
    ):
        self.model = model.to(device)
        self.device = torch.device(device)
        m = unwrap(self.model)
        if max_alpha is not None:
            self.max_alpha = max_alpha
        else:
            cfg = getattr(m, "config", None)
            if (
                cfg is not None
                and hasattr(cfg, "train")
                and hasattr(cfg.train, "alpha_high")
            ):
                self.max_alpha = float(cfg.train.alpha_high)
            else:
                self.max_alpha = 16.0
        self.default_alpha = default_alpha
        self.alpha_policy = alpha_policy
        self.training_status = bool(getattr(m, "training", False))
        self._ref: torch.Tensor | None = None
        self._ref_like: Any = None
        self._target_hw: Tuple[int, int] | None = None
        self._ref_was_bgr: bool = False

        self.model.eval()

    def update_reference(self, ref: Any, landmarks: Any | None = None) -> None:
        ref_t, ref_swapped = _to_tensor(ref, self.device)
        ref_t = _crop_to_multiple(ref_t, multiple=8)
        self._ref = ref_t
        self._ref_like = ref
        self._ref_was_bgr = ref_swapped
        _, h, w = ref_t.shape
        self._target_hw = (h, w)

    def _forward_pair(
        self,
        ref: torch.Tensor,
        frame: torch.Tensor,
        *,
        alpha: float,
        mask: torch.Tensor | None,
    ) -> torch.Tensor:
        our_alpha, num_recursions = self.alpha_policy(alpha, self.max_alpha)
        frames = torch.stack([ref.unsqueeze(0), frame.unsqueeze(0)], dim=2)
        preds = None
        for _ in range(num_recursions):
            out = self.model(frames, alpha=our_alpha, mask=mask)
            preds = out[0] if isinstance(out, (tuple, list)) else out
            frames = torch.stack([ref.unsqueeze(0), preds[0, :, 0].unsqueeze(0)], dim=2)
        return preds[0, :, 0]

    def __call__(
        self,
        frame: Any,
        *,
        alpha: float | None = None,
        mask: Any | None = None,
        landmarks: Any | None = None,
        depth: Any | None = None,
        return_info: bool = False,
    ) -> Any:
        if self._ref is None:
            self.update_reference(frame, landmarks)
        frame_t, frame_swapped = _to_tensor(frame, self.device)
        ref_t = self._ref
        if ref_t is None:
            raise RuntimeError("Reference frame is not set.")

        frame_t = _crop_to_multiple(frame_t, multiple=8, target_hw=self._target_hw)
        ref_t = _crop_to_multiple(ref_t, multiple=8, target_hw=self._target_hw)
        self._ref = ref_t

        mask_t: torch.Tensor | None = None
        if mask is not None:
            mask_t, _ = _to_tensor(mask, self.device)
            mask_t = _crop_to_multiple(mask_t, multiple=8, target_hw=self._target_hw)
            if mask_t.ndim == 3 and mask_t.shape[0] != 1:
                mask_t = mask_t[0:1]

        used_alpha = self.default_alpha if alpha is None else alpha
        with torch.no_grad():
            pred = self._forward_pair(ref_t, frame_t, alpha=used_alpha, mask=mask_t)

        output = _to_numpy(
            pred,
            self._ref_like if self._ref_like is not None else frame,
            swap_to_bgr=self._ref_was_bgr or frame_swapped,
        )
        if return_info:
            info = {
                "alpha": used_alpha,
                "max_alpha": self.max_alpha,
                "device": str(self.device),
            }
            return output, info
        return output


@dataclass
class TTAStep:
    loss: float
    info: Dict[str, float]


class FlowMagTTA:
    """Minimal test-time adaptation helper, separated from I/O concerns."""

    def __init__(
        self,
        model: nn.Module,
        *,
        lr: float = 1e-4,
        device: str | torch.device = "cuda",
        optimizer_factory: Callable[[Iterable[torch.Tensor]], torch.optim.Optimizer]
        | None = None,
    ):
        self.model = model.to(device)
        self.device = torch.device(device)
        m = unwrap(self.model)
        params = (
            m.trainable_parameters()
            if hasattr(m, "trainable_parameters")
            else self.model.parameters()
        )
        self.opt = (
            optimizer_factory(params)
            if optimizer_factory is not None
            else torch.optim.Adam(params, lr=lr)
        )

    def step(self, batch: torch.Tensor, *, alpha: float | None = None) -> TTAStep:
        self.model.train()
        batch = batch.to(self.device)
        out = self.model(batch, alpha=alpha)
        if isinstance(out, (tuple, list)):
            _, loss, info = out
        else:
            raise RuntimeError("Model did not return loss tuple in TTA mode.")

        loss = loss.mean()
        self.opt.zero_grad(set_to_none=True)
        loss.backward()
        self.opt.step()

        info_f = {k: float(v.detach().mean().cpu()) for k, v in (info or {}).items()}
        return TTAStep(loss=float(loss.detach().cpu()), info=info_f)


__all__ = [
    "OnlineMagnifier",
    "FlowMagOnline",
    "FlowMagTTA",
    "TTAStep",
    "unwrap",
    "default_alpha_policy",
    "load_flowmag_model",
]



import collections


def update_dict(orig_dict, new_dict):
    for key, val in new_dict.items():
        if isinstance(val, collections.Mapping):
            tmp = update_dict(orig_dict.get(key, {}), val)
            orig_dict[key] = tmp
        else:
            orig_dict[key] = val
    return orig_dict


class AverageMeter(object):
    """Computes and stores the average and current value"""

    def __init__(self, i=1, precision=3, names=None):
        self.meters = i
        self.precision = precision
        self.reset(self.meters)
        self.names = names
        if names is not None:
            assert self.meters == len(self.names)
        else:
            self.names = [""] * self.meters

    def reset(self, i):
        self.val = [0] * i
        self.avg = [0] * i
        self.sum = [0] * i
        self.count = [0] * i

    def update(self, val, n=1):
        if not isinstance(val, list):
            val = [val]
        if not isinstance(n, list):
            n = [n] * self.meters
        assert len(val) == self.meters and len(n) == self.meters
        for i in range(self.meters):
            self.count[i] += n[i]
        for i, v in enumerate(val):
            self.val[i] = v
            self.sum[i] += v * n[i]
            self.avg[i] = self.sum[i] / self.count[i]

    def __repr__(self):
        val = " ".join(
            [
                "{} {:.{}f}".format(n, v, self.precision)
                for n, v in zip(self.names, self.val)
            ]
        )
        avg = " ".join(
            [
                "{} {:.{}f}".format(n, a, self.precision)
                for n, a in zip(self.names, self.avg)
            ]
        )
        return "{} ({})".format(val, avg)



import os
import cv2
import tqdm
import glob
import argparse

if __name__ == "__main__":
    # save videos in uvo_videos_dense as image files
    parser = argparse.ArgumentParser(description="Process some inputs.")
    parser.add_argument("--data_root", type=str, help="root folder to save data")
    args = parser.parse_args()
    data_root = args.data_root

    path = os.path.join(data_root, "uvo")
    save_root = os.path.join(data_root, "uvo_frames")
    split = "uvo_videos_dense"

    all_videos = glob.glob(os.path.join(path, split, "*.mp4"))
    for video in tqdm.tqdm(all_videos):
        fn = os.path.basename(video).split(".mp4")[0]
        os.makedirs(os.path.join(save_root, split, fn))
        cap = cv2.VideoCapture(video)
        ret, frame = cap.read()
        i = 0
        while ret:
            cv2.imwrite(
                os.path.join(save_root, split, fn, "{}.png".format(str(i).zfill(5))),
                frame,
            )
            i += 1
            ret, frame = cap.read()
        cap.release()



import torch
import torch.nn.functional as F


def coords_grid(b, h, w, homogeneous=False, device=None):
    y, x = torch.meshgrid(torch.arange(h), torch.arange(w))  # [H, W]

    stacks = [x, y]

    if homogeneous:
        ones = torch.ones_like(x)  # [H, W]
        stacks.append(ones)

    grid = torch.stack(stacks, dim=0).float()  # [2, H, W] or [3, H, W]

    grid = grid[None].repeat(b, 1, 1, 1)  # [B, 2, H, W] or [B, 3, H, W]

    if device is not None:
        grid = grid.to(device)

    return grid


def generate_window_grid(h_min, h_max, w_min, w_max, len_h, len_w, device=None):
    assert device is not None

    x, y = torch.meshgrid(
        [
            torch.linspace(w_min, w_max, len_w, device=device),
            torch.linspace(h_min, h_max, len_h, device=device),
        ],
    )
    grid = torch.stack((x, y), -1).transpose(0, 1).float()  # [H, W, 2]

    return grid


def normalize_coords(coords, h, w):
    # coords: [B, H, W, 2]
    c = torch.Tensor([(w - 1) / 2.0, (h - 1) / 2.0]).float().to(coords.device)
    return (coords - c) / c  # [-1, 1]


def bilinear_sample(
    img, sample_coords, mode="bilinear", padding_mode="zeros", return_mask=False
):
    # img: [B, C, H, W]
    # sample_coords: [B, 2, H, W] in image scale
    if sample_coords.size(1) != 2:  # [B, H, W, 2]
        sample_coords = sample_coords.permute(0, 3, 1, 2)

    b, _, h, w = sample_coords.shape

    # Normalize to [-1, 1]
    x_grid = 2 * sample_coords[:, 0] / (w - 1) - 1
    y_grid = 2 * sample_coords[:, 1] / (h - 1) - 1

    grid = torch.stack([x_grid, y_grid], dim=-1)  # [B, H, W, 2]

    img = F.grid_sample(
        img, grid, mode=mode, padding_mode=padding_mode, align_corners=True
    )

    if return_mask:
        mask = (
            (x_grid >= -1) & (y_grid >= -1) & (x_grid <= 1) & (y_grid <= 1)
        )  # [B, H, W]

        return img, mask

    return img


def flow_warp(feature, flow, mask=False, padding_mode="zeros"):
    b, c, h, w = feature.size()
    assert flow.size(1) == 2

    grid = coords_grid(b, h, w).to(flow.device) + flow  # [B, 2, H, W]

    return bilinear_sample(feature, grid, padding_mode=padding_mode, return_mask=mask)


def forward_backward_consistency_check(fwd_flow, bwd_flow, alpha=0.01, beta=0.5):
    # fwd_flow, bwd_flow: [B, 2, H, W]
    # alpha and beta values are following UnFlow (https://arxiv.org/abs/1711.07837)
    assert fwd_flow.dim() == 4 and bwd_flow.dim() == 4
    assert fwd_flow.size(1) == 2 and bwd_flow.size(1) == 2
    flow_mag = torch.norm(fwd_flow, dim=1) + torch.norm(bwd_flow, dim=1)  # [B, H, W]

    warped_bwd_flow = flow_warp(bwd_flow, fwd_flow)  # [B, 2, H, W]
    warped_fwd_flow = flow_warp(fwd_flow, bwd_flow)  # [B, 2, H, W]

    diff_fwd = torch.norm(fwd_flow + warped_bwd_flow, dim=1)  # [B, H, W]
    diff_bwd = torch.norm(bwd_flow + warped_fwd_flow, dim=1)

    threshold = alpha * flow_mag + beta

    fwd_occ = (diff_fwd > threshold).float()  # [B, H, W]
    bwd_occ = (diff_bwd > threshold).float()

    return fwd_occ, bwd_occ



from . import tfw_util as _tfw_util

HAS_YOLOV5_FACE = _tfw_util.HAS_YOLOV5_FACE
HAS_GDOWN = _tfw_util.HAS_GDOWN

__all__ = [
    "HAS_YOLOV5_FACE",
    "HAS_GDOWN",
    *_tfw_util.__all__,
]

for _name in _tfw_util.__all__:
    globals()[_name] = getattr(_tfw_util, _name)

del _tfw_util, _name



import os
import cv2
import json
import tqdm
import shutil
from PIL import Image
from torchvision import transforms
from torch.utils.data import Dataset, DataLoader


class ResizeSet(Dataset):
    def __init__(self, data_root, split, img_size=512):
        self.data_root = data_root

        self.transforms = transforms.Compose(
            [transforms.Resize(img_size), transforms.ToTensor()]
        )
        self.frameA_root = os.path.join(self.data_root, split, "frameA")
        self.frameB_root = os.path.join(self.data_root, split, "frameB")

        with open(os.path.join(self.data_root, "{}_fn.json".format(split)), "r") as f:
            self.fn = json.load(f)

    def __getitem__(self, index):
        image_paths = [
            os.path.join(self.frameA_root, self.fn[index]),
            os.path.join(self.frameB_root, self.fn[index]),
        ]
        images = [self.transforms(Image.open(path)) for path in image_paths]
        return images, self.fn[index]

    def __len__(self):
        return len(self.fn)


def get_loader(data_root, split):
    dataset = ResizeSet(data_root, split)
    return DataLoader(
        dataset, batch_size=1, shuffle=False, num_workers=4, pin_memory=True
    )


def resize_dataset(data_root, split, save_root):
    loader = ResizeSet(data_root, split)

    if not os.path.exists(save_root):
        os.makedirs(save_root)

    if os.path.exists(os.path.join(save_root, split)):
        shutil.rmtree(os.path.join(save_root, split))

    os.makedirs(os.path.join(save_root, split))
    os.makedirs(os.path.join(save_root, split, "frameA"))
    os.makedirs(os.path.join(save_root, split, "frameB"))

    for images, fn in tqdm.tqdm(loader):
        cv2.imwrite(
            os.path.join(save_root, split, "frameA", fn),
            255 * images[0].squeeze().permute(1, 2, 0).flip([-1]).numpy(),
        )
        cv2.imwrite(
            os.path.join(save_root, split, "frameB", fn),
            255 * images[1].squeeze().permute(1, 2, 0).flip([-1]).numpy(),
        )



import numpy as np
import torch

# from skimage.color import rgb2yuv
import cv2
from fast_slic.avx2 import SlicAvx2 as Slic
from skimage.segmentation import slic as sk_slic


def run_slic_pt(
    img_batch, n_seg=200, compact=10, rd_select=(8, 16), fast=True
):  # Nx1xHxW
    """

    :param img: Nx3xHxW 0~1 float32
    :param n_seg:
    :param compact:
    :return: Nx1xHxW float32
    """
    B = img_batch.size(0)
    dtype = img_batch.type()
    img_batch = np.split(
        img_batch.detach().cpu().numpy().transpose([0, 2, 3, 1]), B, axis=0
    )
    out = []
    if fast:
        fast_slic = Slic(num_components=n_seg, compactness=compact, min_size_factor=0.8)
    for img in img_batch:
        img = np.copy((img * 255).squeeze(0).astype(np.uint8), order="C")
        if fast:
            img = cv2.cvtColor(img, cv2.COLOR_RGB2LAB)
            seg = fast_slic.iterate(img)
        else:
            seg = sk_slic(img, n_segments=200, compactness=10)

        if rd_select is not None:
            n_select = np.random.randint(rd_select[0], rd_select[1])
            select_list = np.random.choice(
                range(0, np.max(seg) + 1), n_select, replace=False
            )

            seg = np.bitwise_or.reduce([seg == seg_id for seg_id in select_list])
        out.append(seg)
    x_out = torch.tensor(np.stack(out)).type(dtype).unsqueeze(1)
    return x_out


def random_crop(img, flow, occ_mask, crop_sz):
    """

    :param img: Nx6xHxW
    :param flows: n * [Nx2xHxW]
    :param occ_masks: n * [Nx1xHxW]
    :param crop_sz:
    :return:
    """
    _, _, h, w = img.size()
    c_h, c_w = crop_sz

    if c_h == h and c_w == w:
        return img, flow, occ_mask

    x1 = np.random.randint(0, w - c_w)
    y1 = np.random.randint(0, h - c_h)
    img = img[:, :, y1 : y1 + c_h, x1 : x1 + c_w]
    flow = flow[:, :, y1 : y1 + c_h, x1 : x1 + c_w]
    occ_mask = occ_mask[:, :, y1 : y1 + c_h, x1 : x1 + c_w]

    return img, flow, occ_mask



import numpy as np
import torch
import torch.nn as nn
from einops import repeat, rearrange

from neurovc.contrib.flowmag.models.model_zoo import UNet
from neurovc.contrib.flowmag.losses import MMLoss


class MotionMagModel(nn.Module):
    def __init__(self, config):
        super().__init__()
        self.config = config
        self.generator = UNet(
            n_channels=config.data.num_frames * 3 + config.model.pos_dim,
            n_classes=(config.data.num_frames - 1) * 3,
            num_layers=config.model.num_layers,
            ndf=config.model.ndf,
            final_activation="sigmoid",
        )
        self.alpha_appender = AlphaAppender(config)
        if config.train.is_training:
            self.training = True
            self.mm_loss_fn = MMLoss(config)
        else:
            self.training = False

    def get_training_status(self):
        return True if self.training else False

    def trainable_parameters(self):
        return self.generator.parameters()

    def forward(self, frames, alpha=None, mask=None):
        b, c, t, h, w = frames.shape

        # Predict
        frames_flat = rearrange(frames, "b c t h w -> b (c t) h w")
        frames_flat, alphas = self.alpha_appender(frames_flat, alpha=alpha, mask=mask)
        preds = self.generator(frames_flat)
        preds = rearrange(preds, "b (c t) h w -> b c t h w", c=c, t=t - 1)

        # Compute Losses
        if self.training:
            loss_mag, loss_color, info = self.mm_loss_fn(preds, frames, alphas)
            loss_mag = loss_mag.mean()
            loss_color = loss_color.mean()
            loss = (
                self.config.train.weight_mag * loss_mag
                + self.config.train.weight_color * loss_color
            )

            # Add stuff to info
            info["loss_mag"] = loss_mag
            info["loss_color"] = loss_color

            return preds, loss, info

        else:
            return preds


class AlphaAppender(nn.Module):
    def __init__(self, config):
        super().__init__()
        self.alpha_low = config.train.alpha_low
        self.alpha_high = config.train.alpha_high
        self.pos_dim = config.model.pos_dim

    def sample_alphas(self, b):
        # Sample alpha exponentially with a size of b
        exp_low = np.log2(self.alpha_low)
        exp_high = np.log2(self.alpha_high)
        exps = np.random.uniform(exp_low, exp_high, b)
        alphas = 2**exps
        return torch.tensor(alphas).float()

    def get_positional_encoding(self, alphas):
        """
        Given alphas of shape (batch_dim,), return positional encoding
            of shape (batch_dim, pos_dim)
        """
        device = alphas.device

        # Wavelengths are geometric between 0.125 and 128
        wavelengths = torch.logspace(-3, 7, self.pos_dim, base=2).to(device)
        vals = torch.sin(2 * 3.141592 * alphas[:, None] / wavelengths[None, :])
        return vals

    def forward(self, frames, alpha=None, mask=None):
        b, c, h, w = frames.shape
        device = frames.device

        # Sample alphas
        if alpha is None:
            alphas = self.sample_alphas(b).to(device)
        else:
            alphas = torch.ones(b).to(device) * alpha

        # Get positional encoding
        pe = self.get_positional_encoding(alphas)
        pe = pe.to(frames.device)

        # Repeat spatially and concat
        pe = repeat(pe, "b d -> b d h w", h=h, w=w)

        if mask is not None:
            # Add batch and pos_dim dims
            mask = mask[None, None]

            # Get "zero" motion mag
            zero_alpha = self.get_positional_encoding(torch.ones(b))
            zero_alpha = zero_alpha.to(frames.device)
            zero_alpha = repeat(zero_alpha, "b d -> b d h w", h=h, w=w)

            # Replace zeros with "zero_alpha"
            pe = pe * mask + zero_alpha * (1 - mask)

        # Concatenate frames with alpha embedding
        frames_pe = torch.cat([frames, pe], dim=1)

        # Return both frames_pe and alphas used
        return frames_pe, alphas.to(frames.device)



import numpy as np
import torch

# from scipy.misc import imresize
from skimage.transform import resize as imresize


class ArrayToTensor(object):
    """Converts a numpy.ndarray (H x W x C) to a torch.FloatTensor of shape (C x H x W)."""

    def __call__(self, array):
        assert isinstance(array, np.ndarray)
        array = np.transpose(array, (2, 0, 1))
        # handle numpy array
        tensor = torch.from_numpy(array)
        # put it from HWC to CHW format
        return tensor.float()


class Zoom(object):
    def __init__(self, new_h, new_w):
        self.new_h = new_h
        self.new_w = new_w

    def __call__(self, image):
        h, w, _ = image.shape
        if h == self.new_h and w == self.new_w:
            return image
        image = imresize(image, (self.new_h, self.new_w))
        return image



import torch
import torch.nn as nn
import torch.nn.functional as F

from neurovc.contrib.flowmag.flow_models.gmflow.backbone import CNNEncoder
from neurovc.contrib.flowmag.flow_models.gmflow.geometry import flow_warp
from neurovc.contrib.flowmag.flow_models.gmflow.matching import (
    global_correlation_softmax,
    local_correlation_softmax,
)
from neurovc.contrib.flowmag.flow_models.gmflow.transformer import (
    FeatureFlowAttention,
    FeatureTransformer,
)
from neurovc.contrib.flowmag.flow_models.gmflow.utils import (
    feature_add_position,
    normalize_img,
)


class GMFlow(nn.Module):
    def __init__(
        self,
        num_scales=1,
        upsample_factor=8,
        feature_channels=128,
        attention_type="swin",
        num_transformer_layers=6,
        ffn_dim_expansion=4,
        num_head=1,
        **kwargs,
    ):
        super(GMFlow, self).__init__()

        self.num_scales = num_scales
        self.feature_channels = feature_channels
        self.upsample_factor = upsample_factor
        self.attention_type = attention_type
        self.num_transformer_layers = num_transformer_layers

        # CNN backbone
        self.backbone = CNNEncoder(
            output_dim=feature_channels, num_output_scales=num_scales
        )

        # Transformer
        self.transformer = FeatureTransformer(
            num_layers=num_transformer_layers,
            d_model=feature_channels,
            nhead=num_head,
            attention_type=attention_type,
            ffn_dim_expansion=ffn_dim_expansion,
        )

        # flow propagation with self-attn
        self.feature_flow_attn = FeatureFlowAttention(in_channels=feature_channels)

        # convex upsampling: concat feature0 and flow as input
        self.upsampler = nn.Sequential(
            nn.Conv2d(2 + feature_channels, 256, 3, 1, 1),
            nn.ReLU(inplace=True),
            nn.Conv2d(256, upsample_factor**2 * 9, 1, 1, 0),
        )

    def extract_feature(self, img0, img1):
        concat = torch.cat((img0, img1), dim=0)  # [2B, C, H, W]
        features = self.backbone(
            concat
        )  # list of [2B, C, H, W], resolution from high to low

        # reverse: resolution from low to high
        features = features[::-1]

        feature0, feature1 = [], []

        for i in range(len(features)):
            feature = features[i]
            chunks = torch.chunk(feature, 2, 0)  # tuple
            feature0.append(chunks[0])
            feature1.append(chunks[1])

        return feature0, feature1

    def upsample_flow(
        self,
        flow,
        feature,
        bilinear=False,
        upsample_factor=8,
    ):
        if bilinear:
            up_flow = (
                F.interpolate(
                    flow,
                    scale_factor=upsample_factor,
                    mode="bilinear",
                    align_corners=True,
                )
                * upsample_factor
            )

        else:
            # convex upsampling
            concat = torch.cat((flow, feature), dim=1)

            mask = self.upsampler(concat)
            b, flow_channel, h, w = flow.shape
            mask = mask.view(
                b, 1, 9, self.upsample_factor, self.upsample_factor, h, w
            )  # [B, 1, 9, K, K, H, W]
            mask = torch.softmax(mask, dim=2)

            up_flow = F.unfold(self.upsample_factor * flow, [3, 3], padding=1)
            up_flow = up_flow.view(
                b, flow_channel, 9, 1, 1, h, w
            )  # [B, 2, 9, 1, 1, H, W]

            up_flow = torch.sum(mask * up_flow, dim=2)  # [B, 2, K, K, H, W]
            up_flow = up_flow.permute(0, 1, 4, 2, 5, 3)  # [B, 2, K, H, K, W]
            up_flow = up_flow.reshape(
                b, flow_channel, self.upsample_factor * h, self.upsample_factor * w
            )  # [B, 2, K*H, K*W]

        return up_flow

    def forward(
        self,
        img0,
        img1,
        attn_splits_list=None,
        corr_radius_list=None,
        prop_radius_list=None,
        pred_bidir_flow=False,
        **kwargs,
    ):
        results_dict = {}
        flow_preds = []

        img0, img1 = normalize_img(img0, img1)  # [B, 3, H, W]

        # resolution low to high
        feature0_list, feature1_list = self.extract_feature(
            img0, img1
        )  # list of features

        flow = None

        assert (
            len(attn_splits_list)
            == len(corr_radius_list)
            == len(prop_radius_list)
            == self.num_scales
        )

        for scale_idx in range(self.num_scales):
            feature0, feature1 = feature0_list[scale_idx], feature1_list[scale_idx]

            if pred_bidir_flow and scale_idx > 0:
                # predicting bidirectional flow with refinement
                feature0, feature1 = (
                    torch.cat((feature0, feature1), dim=0),
                    torch.cat((feature1, feature0), dim=0),
                )

            upsample_factor = self.upsample_factor * (
                2 ** (self.num_scales - 1 - scale_idx)
            )

            if scale_idx > 0:
                flow = (
                    F.interpolate(
                        flow, scale_factor=2, mode="bilinear", align_corners=True
                    )
                    * 2
                )

            if flow is not None:
                flow = flow.detach()
                feature1 = flow_warp(feature1, flow)  # [B, C, H, W]

            attn_splits = attn_splits_list[scale_idx]
            corr_radius = corr_radius_list[scale_idx]
            prop_radius = prop_radius_list[scale_idx]

            # add position to features
            feature0, feature1 = feature_add_position(
                feature0, feature1, attn_splits, self.feature_channels
            )

            # Transformer
            feature0, feature1 = self.transformer(
                feature0, feature1, attn_num_splits=attn_splits
            )

            # correlation and softmax
            if corr_radius == -1:  # global matching
                flow_pred = global_correlation_softmax(
                    feature0, feature1, pred_bidir_flow
                )[0]
            else:  # local matching
                flow_pred = local_correlation_softmax(feature0, feature1, corr_radius)[
                    0
                ]

            # flow or residual flow
            flow = flow + flow_pred if flow is not None else flow_pred

            # upsample to the original resolution for supervison
            if (
                self.training
            ):  # only need to upsample intermediate flow predictions at training time
                flow_bilinear = self.upsample_flow(
                    flow, None, bilinear=True, upsample_factor=upsample_factor
                )
                flow_preds.append(flow_bilinear)

            # flow propagation with self-attn
            if pred_bidir_flow and scale_idx == 0:
                feature0 = torch.cat(
                    (feature0, feature1), dim=0
                )  # [2*B, C, H, W] for propagation
            flow = self.feature_flow_attn(
                feature0,
                flow.detach(),
                local_window_attn=prop_radius > 0,
                local_window_radius=prop_radius,
            )

            # bilinear upsampling at training time except the last one
            if self.training and scale_idx < self.num_scales - 1:
                flow_up = self.upsample_flow(
                    flow, feature0, bilinear=True, upsample_factor=upsample_factor
                )
                flow_preds.append(flow_up)

            if scale_idx == self.num_scales - 1:
                flow_up = self.upsample_flow(flow, feature0)
                flow_preds.append(flow_up)

        results_dict.update({"flow_preds": flow_preds})

        return results_dict



import torch
from utils.torch_utils import init_seed

from datasets.get_dataset import get_dataset
from models.get_model import get_model
from losses.get_loss import get_loss
from trainer.get_trainer import get_trainer


def main(cfg, _log):
    init_seed(cfg.seed)

    _log.info("=> fetching img pairs.")
    train_set, valid_set = get_dataset(cfg)

    _log.info(
        "{} samples found, {} train samples and {} test samples ".format(
            len(valid_set) + len(train_set), len(train_set), len(valid_set)
        )
    )

    train_loader = torch.utils.data.DataLoader(
        train_set,
        batch_size=cfg.train.batch_size,
        num_workers=cfg.train.workers,
        pin_memory=True,
        shuffle=True,
    )

    max_test_batch = 4
    if type(valid_set) is torch.utils.data.ConcatDataset:
        valid_loader = [
            torch.utils.data.DataLoader(
                s,
                batch_size=min(max_test_batch, cfg.train.batch_size),
                num_workers=min(4, cfg.train.workers),
                pin_memory=True,
                shuffle=False,
            )
            for s in valid_set.datasets
        ]
        valid_size = sum(len(loader) for loader in valid_loader)
    else:
        valid_loader = torch.utils.data.DataLoader(
            valid_set,
            batch_size=min(max_test_batch, cfg.train.batch_size),
            num_workers=min(4, cfg.train.workers),
            pin_memory=True,
            shuffle=False,
        )
        valid_size = len(valid_loader)

    if cfg.train.epoch_size == 0:
        cfg.train.epoch_size = len(train_loader)
    if cfg.train.valid_size == 0:
        cfg.train.valid_size = valid_size
    cfg.train.epoch_size = min(cfg.train.epoch_size, len(train_loader))
    cfg.train.valid_size = min(cfg.train.valid_size, valid_size)

    model = get_model(cfg.model)
    loss = get_loss(cfg.loss)
    trainer = get_trainer(cfg.trainer)(
        train_loader, valid_loader, model, loss, _log, cfg.save_root, cfg.train
    )

    trainer.train()



# Data loading based on https://github.com/NVIDIA/flownet2-pytorch

import numpy as np
import torch
import torch.utils.data as data

import os
import random
from glob import glob
import os.path as osp

from raft_utils import frame_utils
from raft_utils.augmentor import FlowAugmentor, SparseFlowAugmentor


class FlowDataset(data.Dataset):
    def __init__(self, aug_params=None, sparse=False):
        self.augmentor = None
        self.sparse = sparse
        if aug_params is not None:
            if sparse:
                self.augmentor = SparseFlowAugmentor(**aug_params)
            else:
                self.augmentor = FlowAugmentor(**aug_params)

        self.is_test = False
        self.init_seed = False
        self.flow_list = []
        self.image_list = []
        self.extra_info = []

    def __getitem__(self, index):
        if self.is_test:
            img1 = frame_utils.read_gen(self.image_list[index][0])
            img2 = frame_utils.read_gen(self.image_list[index][1])
            img1 = np.array(img1).astype(np.uint8)[..., :3]
            img2 = np.array(img2).astype(np.uint8)[..., :3]
            img1 = torch.from_numpy(img1).permute(2, 0, 1).float()
            img2 = torch.from_numpy(img2).permute(2, 0, 1).float()
            return img1, img2, self.extra_info[index]

        if not self.init_seed:
            worker_info = torch.utils.data.get_worker_info()
            if worker_info is not None:
                torch.manual_seed(worker_info.id)
                np.random.seed(worker_info.id)
                random.seed(worker_info.id)
                self.init_seed = True

        index = index % len(self.image_list)
        valid = None
        if self.sparse:
            flow, valid = frame_utils.readFlowKITTI(self.flow_list[index])
        else:
            flow = frame_utils.read_gen(self.flow_list[index])

        img1 = frame_utils.read_gen(self.image_list[index][0])
        img2 = frame_utils.read_gen(self.image_list[index][1])

        flow = np.array(flow).astype(np.float32)
        img1 = np.array(img1).astype(np.uint8)
        img2 = np.array(img2).astype(np.uint8)

        # grayscale images
        if len(img1.shape) == 2:
            img1 = np.tile(img1[..., None], (1, 1, 3))
            img2 = np.tile(img2[..., None], (1, 1, 3))
        else:
            img1 = img1[..., :3]
            img2 = img2[..., :3]

        if self.augmentor is not None:
            if self.sparse:
                img1, img2, flow, valid = self.augmentor(img1, img2, flow, valid)
            else:
                img1, img2, flow = self.augmentor(img1, img2, flow)

        img1 = torch.from_numpy(img1).permute(2, 0, 1).float()
        img2 = torch.from_numpy(img2).permute(2, 0, 1).float()
        flow = torch.from_numpy(flow).permute(2, 0, 1).float()

        if valid is not None:
            valid = torch.from_numpy(valid)
        else:
            valid = (flow[0].abs() < 1000) & (flow[1].abs() < 1000)

        return img1, img2, flow, valid.float()

    def __rmul__(self, v):
        self.flow_list = v * self.flow_list
        self.image_list = v * self.image_list
        return self

    def __len__(self):
        return len(self.image_list)


class MpiSintel(FlowDataset):
    def __init__(
        self, aug_params=None, split="training", root="datasets/Sintel", dstype="clean"
    ):
        super(MpiSintel, self).__init__(aug_params)
        flow_root = osp.join(root, split, "flow")
        image_root = osp.join(root, split, dstype)

        if split == "test":
            self.is_test = True

        for scene in os.listdir(image_root):
            image_list = sorted(glob(osp.join(image_root, scene, "*.png")))
            for i in range(len(image_list) - 1):
                self.image_list += [[image_list[i], image_list[i + 1]]]
                self.extra_info += [(scene, i)]  # scene and frame_id

            if split != "test":
                self.flow_list += sorted(glob(osp.join(flow_root, scene, "*.flo")))


class FlyingChairs(FlowDataset):
    def __init__(
        self, aug_params=None, split="train", root="datasets/FlyingChairs_release/data"
    ):
        super(FlyingChairs, self).__init__(aug_params)

        images = sorted(glob(osp.join(root, "*.ppm")))
        flows = sorted(glob(osp.join(root, "*.flo")))
        assert len(images) // 2 == len(flows)

        split_list = np.loadtxt("chairs_split.txt", dtype=np.int32)
        for i in range(len(flows)):
            xid = split_list[i]
            if (split == "training" and xid == 1) or (
                split == "validation" and xid == 2
            ):
                self.flow_list += [flows[i]]
                self.image_list += [[images[2 * i], images[2 * i + 1]]]


class FlyingThings3D(FlowDataset):
    def __init__(
        self, aug_params=None, root="datasets/FlyingThings3D", dstype="frames_cleanpass"
    ):
        super(FlyingThings3D, self).__init__(aug_params)

        for cam in ["left"]:
            for direction in ["into_future", "into_past"]:
                image_dirs = sorted(glob(osp.join(root, dstype, "TRAIN/*/*")))
                image_dirs = sorted([osp.join(f, cam) for f in image_dirs])

                flow_dirs = sorted(glob(osp.join(root, "optical_flow/TRAIN/*/*")))
                flow_dirs = sorted([osp.join(f, direction, cam) for f in flow_dirs])

                for idir, fdir in zip(image_dirs, flow_dirs):
                    images = sorted(glob(osp.join(idir, "*.png")))
                    flows = sorted(glob(osp.join(fdir, "*.pfm")))
                    for i in range(len(flows) - 1):
                        if direction == "into_future":
                            self.image_list += [[images[i], images[i + 1]]]
                            self.flow_list += [flows[i]]
                        elif direction == "into_past":
                            self.image_list += [[images[i + 1], images[i]]]
                            self.flow_list += [flows[i + 1]]


class KITTI(FlowDataset):
    def __init__(self, aug_params=None, split="training", root="datasets/KITTI"):
        super(KITTI, self).__init__(aug_params, sparse=True)
        if split == "testing":
            self.is_test = True

        root = osp.join(root, split)
        images1 = sorted(glob(osp.join(root, "image_2/*_10.png")))
        images2 = sorted(glob(osp.join(root, "image_2/*_11.png")))

        for img1, img2 in zip(images1, images2):
            frame_id = img1.split("/")[-1]
            self.extra_info += [[frame_id]]
            self.image_list += [[img1, img2]]

        if split == "training":
            self.flow_list = sorted(glob(osp.join(root, "flow_occ/*_10.png")))


class HD1K(FlowDataset):
    def __init__(self, aug_params=None, root="datasets/HD1k"):
        super(HD1K, self).__init__(aug_params, sparse=True)

        seq_ix = 0
        while 1:
            flows = sorted(
                glob(os.path.join(root, "hd1k_flow_gt", "flow_occ/%06d_*.png" % seq_ix))
            )
            images = sorted(
                glob(os.path.join(root, "hd1k_input", "image_2/%06d_*.png" % seq_ix))
            )

            if len(flows) == 0:
                break

            for i in range(len(flows) - 1):
                self.flow_list += [flows[i]]
                self.image_list += [[images[i], images[i + 1]]]

            seq_ix += 1


def fetch_dataloader(args, TRAIN_DS="C+T+K+S+H"):
    """Create the data loader for the corresponding trainign set"""

    if args.stage == "chairs":
        aug_params = {
            "crop_size": args.image_size,
            "min_scale": -0.1,
            "max_scale": 1.0,
            "do_flip": True,
        }
        train_dataset = FlyingChairs(aug_params, split="training")

    elif args.stage == "things":
        aug_params = {
            "crop_size": args.image_size,
            "min_scale": -0.4,
            "max_scale": 0.8,
            "do_flip": True,
        }
        clean_dataset = FlyingThings3D(aug_params, dstype="frames_cleanpass")
        final_dataset = FlyingThings3D(aug_params, dstype="frames_finalpass")
        train_dataset = clean_dataset + final_dataset

    elif args.stage == "sintel":
        aug_params = {
            "crop_size": args.image_size,
            "min_scale": -0.2,
            "max_scale": 0.6,
            "do_flip": True,
        }
        things = FlyingThings3D(aug_params, dstype="frames_cleanpass")
        sintel_clean = MpiSintel(aug_params, split="training", dstype="clean")
        sintel_final = MpiSintel(aug_params, split="training", dstype="final")

        if TRAIN_DS == "C+T+K+S+H":
            kitti = KITTI(
                {
                    "crop_size": args.image_size,
                    "min_scale": -0.3,
                    "max_scale": 0.5,
                    "do_flip": True,
                }
            )
            hd1k = HD1K(
                {
                    "crop_size": args.image_size,
                    "min_scale": -0.5,
                    "max_scale": 0.2,
                    "do_flip": True,
                }
            )
            train_dataset = (
                100 * sintel_clean
                + 100 * sintel_final
                + 200 * kitti
                + 5 * hd1k
                + things
            )

        elif TRAIN_DS == "C+T+K/S":
            train_dataset = 100 * sintel_clean + 100 * sintel_final + things

    elif args.stage == "kitti":
        aug_params = {
            "crop_size": args.image_size,
            "min_scale": -0.2,
            "max_scale": 0.4,
            "do_flip": False,
        }
        train_dataset = KITTI(aug_params, split="training")

    train_loader = data.DataLoader(
        train_dataset,
        batch_size=args.batch_size,
        pin_memory=False,
        shuffle=True,
        num_workers=4,
        drop_last=True,
    )

    print("Training with %d image pairs" % len(train_dataset))
    return train_loader



import numpy as np
import scipy.io
import scipy.ndimage
import aolib.img as ig
import aolib.util as ut
import os


def warp_flow(im2, vx, vy):
    """Assumes im1[y, x] = im2[y + flow[y, x, 1], x + flow[y, x, 0]]"""

    # this code is confusing because we assume vx and vy are the negative
    # of where to send each pixel, as in the results by ce's siftflow code
    y, x = np.mgrid[: im2.shape[0], : im2.shape[1]]
    dy = (y + vy).flatten()[np.newaxis, :]
    dx = (x + vx).flatten()[np.newaxis, :]
    # this says: a recipe for making im1 is to make a new image where im[y, x] = im2[y + flow[y, x, 1], x + flow[y, x, 0]]
    return np.concatenate(
        [
            scipy.ndimage.map_coordinates(
                im2[..., i], np.concatenate([dy, dx])
            ).reshape(im2.shape[:2] + (1,))
            for i in range(im2.shape[2])
        ],
        axis=2,
    )


# given a flow that warps im2 -> im1, returns a warped version that warps im1 to im2
def invert_warp(im1, vx, vy):
    y, x = np.mgrid[: im1.shape[0], : im1.shape[1]]
    im_coords = np.concatenate([x[..., np.newaxis], y[..., np.newaxis]], axis=2)
    new_flow = warp_flow(im_coords, vx, vy) - im_coords
    return warp_flow(im1, -new_flow[..., 0], -new_flow[..., 1])


def test_invert_warp():
    im1 = ig.draw_rects(ig.make(200, 200), [(100, 50, 100, 100)], fills=(0, 0, 255))
    im2 = ig.draw_rects(ig.make(200, 200), [(50, 50, 100, 100)], fills=(255, 0, 0))
    vx = -50 + np.zeros(im1.shape[:2])
    vy = np.zeros(im1.shape[:2])
    ig.show([[im1, im2], warp_flow(im2, vx, vy), invert_warp(im1, vx, vy)])


# def color_flow(flow, max_flow = None, show = True):
def color_flow(flow, max_flow=-1, show=True):
    flow = flow.copy()
    # based on flowToColor.m by Deqing Sun, orignally based on code by Daniel Scharstein
    UNKNOWN_FLOW_THRESH = 1e9
    height, width, nbands = flow.shape
    assert nbands == 2
    u, v = flow[:, :, 0], flow[:, :, 1]
    maxu = -999.0
    maxv = -999.0
    minu = 999.0
    minv = 999.0
    maxrad = -1.0

    idx_unknown = ut.lor(
        np.abs(u) > UNKNOWN_FLOW_THRESH, np.abs(v) > UNKNOWN_FLOW_THRESH
    )
    u[idx_unknown] = 0
    v[idx_unknown] = 0

    maxu = max(maxu, np.max(u))
    maxv = max(maxv, np.max(v))

    minu = min(minu, np.min(u))
    minv = min(minv, np.min(v))

    rad = np.sqrt(u**2 + v**2)
    maxrad = max(maxrad, np.max(rad))

    if show:
        print(
            "max flow: %.4f flow range: u = %.3f .. %.3f; v = %.3f .. %.3f\n"
            % (maxrad, minu, maxu, minv, maxv)
        )

    if max_flow > 0:
        maxrad = max_flow

    u = u / (maxrad + np.spacing(1))
    v = v / (maxrad + np.spacing(1))

    im = compute_color(u, v)
    im[idx_unknown] = 0
    return im


def compute_color(u, v):
    # from same source as color_flow; please see above comment
    nan_idx = ut.lor(np.isnan(u), np.isnan(v))
    u[nan_idx] = 0
    v[nan_idx] = 0
    colorwheel = make_color_wheel()
    ncols = colorwheel.shape[0]

    rad = np.sqrt(u**2 + v**2)

    a = np.arctan2(-v, -u) / np.pi

    # fk = (a + 1)/2. * (ncols-1) + 1
    fk = (a + 1) / 2.0 * (ncols - 1)

    k0 = np.array(np.floor(fk), "l")

    k1 = k0 + 1
    k1[k1 == ncols] = 1

    f = fk - k0

    im = np.zeros(u.shape + (3,))

    for i in range(colorwheel.shape[1]):
        tmp = colorwheel[:, i]
        col0 = tmp[k0] / 255.0
        col1 = tmp[k1] / 255.0
        col = (1 - f) * col0 + f * col1

        idx = rad <= 1
        col[idx] = 1 - rad[idx] * (1 - col[idx])
        col[ut.lnot(idx)] *= 0.75
        im[:, :, i] = np.uint8(np.floor(255 * col * (1 - nan_idx)))

    return im


def make_color_wheel():
    # same source as color_flow

    RY = 15
    YG = 6
    GC = 4
    CB = 11
    BM = 13
    MR = 6

    ncols = RY + YG + GC + CB + BM + MR

    # colorwheel = zeros(ncols, 3) # r g b
    # matlab correction
    colorwheel = np.zeros((1 + ncols, 4))  # r g b

    col = 0
    # RY
    colorwheel[1 : 1 + RY, 1] = 255
    colorwheel[1 : 1 + RY, 2] = np.floor(255 * np.arange(0, 1 + RY - 1) / RY).T
    col = col + RY

    # YG
    colorwheel[col + 1 : col + 1 + YG, 1] = (
        255 - np.floor(255 * np.arange(0, 1 + YG - 1) / YG).T
    )
    colorwheel[col + 1 : col + 1 + YG, 2] = 255
    col = col + YG

    # GC
    colorwheel[col + 1 : col + 1 + GC, 2] = 255
    colorwheel[col + 1 : col + 1 + GC, 3] = np.floor(
        255 * np.arange(0, 1 + GC - 1) / GC
    ).T
    col = col + GC

    # CB
    colorwheel[col + 1 : col + 1 + CB, 2] = (
        255 - np.floor(255 * np.arange(0, 1 + CB - 1) / CB).T
    )
    colorwheel[col + 1 : col + 1 + CB, 3] = 255
    col = col + CB

    # BM
    colorwheel[col + 1 : col + 1 + BM, 3] = 255
    colorwheel[col + 1 : col + 1 + BM, 1] = np.floor(
        255 * np.arange(0, 1 + BM - 1) / BM
    ).T
    col = col + BM

    # MR
    colorwheel[col + 1 : col + 1 + MR, 3] = (
        255 - np.floor(255 * np.arange(0, 1 + MR - 1) / MR).T
    )
    colorwheel[col + 1 : col + 1 + MR, 1] = 255

    # 1-based to 0-based indices
    return colorwheel[1:, 1:]


def ldof(im1, im2, max_dim=None, rescale=True):
    im1_shape = im1.shape
    with ut.temp_file(".ppm") as fname1, ut.temp_file(".ppm") as fname2:
        if max_dim is not None:
            s = max_dim / float(max(im1.shape[:2]))
            im1 = ig.scale(im1, s)
            im2 = ig.scale(im2, s)
        else:
            s = 1.0
        ig.save(fname1, im1)
        ig.save(fname2, im2)
        ut.sys_check('../lib/ldof/ldof "%s" "%s"' % (fname1, fname2))
        out_fname = fname1.replace(".ppm", "LDOF.flo")
        flo = np.array(np.fromfile(out_fname, np.float32), "float32")[3:].reshape(
            (im1.shape[0], im1.shape[1], 2)
        )
        os.remove(out_fname)
        if rescale and max_dim is not None:
            flo = (1.0 / s) * ig.scale(flo, im1_shape)
        return flo


def deepflow(im1, im2, max_dim=None, rescale=True, match=False, shape=None):
    if max_dim is not None:
        im1_shape = im1.shape
    with (
        ut.temp_file(".png") as fname1,
        ut.temp_file(".png") as fname2,
        ut.temp_file(".flo") as flow_file,
    ):
        if max_dim is not None:
            s = max_dim / float(max(im1.shape[:2]))
            im1 = ig.scale(im1, s)
            im2 = ig.scale(im2, s)
        else:
            s = 1.0

        if not isinstance(im1, str) or not isinstance(im2, str):
            ig.save(fname1, im1)
            ig.save(fname2, im2)
            new_shape = im1.shape
        else:
            fname1, fname2 = im1, im2
            new_shape = shape

        if match:
            # ut.sys_check(('../lib/deepmatching_1.2.2_c++/deepmatching-static "%s" "%s" | '
            #               '../lib/DeepFlow_release2.0/deepflow2-static "%s" "%s" "%s" -match') \
            #              % (fname1, fname2, fname1, fname2, flow_file))
            ut.sys_check(
                (
                    '../lib/deepmatching_1.2.2_c++/deepmatching-static "%s" "%s" -downscale 2 | '
                    '../lib/DeepFlow_release2.0/deepflow2-static "%s" "%s" "%s" -match'
                )
                % (fname1, fname2, fname1, fname2, flow_file)
            )
        else:
            ut.sys_check(
                '../lib/DeepFlow_release2.0/deepflow2-static "%s" "%s" "%s"'
                % (fname1, fname2, flow_file)
            )

        flo = np.array(np.fromfile(flow_file, np.float32), "float32")[3:].reshape(
            (new_shape[0], new_shape[1], 2)
        )
        if rescale and max_dim is not None:
            flo = (1.0 / s) * ig.scale(flo, im1_shape)
        return flo


def flow_im(im, flo, dim, scale=2.5):
    # Similar to the representation in Donahue et. al.'s LRCN paper
    flo_scaled = flo * (dim / float(flo.shape[0]))
    flo_scaled = ig.scale(flo_scaled, (dim, dim))
    flo_byte = np.uint8(np.clip((flo_scaled * scale) + 128, 0, 255))
    # Make the last channel be a grayscale version of the image itself
    both = np.zeros((dim, dim, 3))
    both[:, :, :2] = flo_byte
    both[:, :, 2] = ig.scale(ig.luminance(im), (dim, dim))
    return np.uint8(both)



#!/usr/bin/env python

import torch

import getopt
import math
import numpy
import PIL
import PIL.Image
import sys
import numpy as np

try:
    from neurovc.contrib.flowmag.flow_models.pwcnet.pytorch_pwc.correlation import (
        correlation,
    )  # the custom cost volume layer
except ImportError:
    sys.path.insert(0, "./correlation")
    import correlation  # you should consider upgrading python
# end

##########################################################

assert (
    int(str("").join(torch.__version__.split(".")[0:2])) >= 13
)  # requires at least pytorch version 1.3.0

torch.set_grad_enabled(
    False
)  # make sure to not compute gradients for computational performance

torch.backends.cudnn.enabled = (
    True  # make sure to use cudnn for computational performance
)

##########################################################

arguments_strModel = "default"  # 'default', or 'chairs-things'
arguments_strOut = "./out.flo"

for strOption, strArgument in getopt.getopt(
    sys.argv[1:], "", [strParameter[2:] + "=" for strParameter in sys.argv[1::2]]
)[0]:
    if strOption == "--model" and strArgument != "":
        arguments_strModel = strArgument  # which model to use
    if strOption == "--pairs" and strArgument != "":
        arguments_strModel = strArgument  # which model to use
    if strOption == "--out" and strArgument != "":
        arguments_strOut = strArgument  # path to where the output should be stored
# end

##########################################################

backwarp_tenGrid = {}
backwarp_tenPartial = {}


def backwarp(tenInput, tenFlow):
    if str(tenFlow.shape) not in backwarp_tenGrid:
        tenHor = (
            torch.linspace(
                -1.0 + (1.0 / tenFlow.shape[3]),
                1.0 - (1.0 / tenFlow.shape[3]),
                tenFlow.shape[3],
            )
            .view(1, 1, 1, -1)
            .expand(-1, -1, tenFlow.shape[2], -1)
        )
        tenVer = (
            torch.linspace(
                -1.0 + (1.0 / tenFlow.shape[2]),
                1.0 - (1.0 / tenFlow.shape[2]),
                tenFlow.shape[2],
            )
            .view(1, 1, -1, 1)
            .expand(-1, -1, -1, tenFlow.shape[3])
        )

        backwarp_tenGrid[str(tenFlow.shape)] = torch.cat([tenHor, tenVer], 1).cuda()
    # end

    if str(tenFlow.shape) not in backwarp_tenPartial:
        backwarp_tenPartial[str(tenFlow.shape)] = tenFlow.new_ones(
            [tenFlow.shape[0], 1, tenFlow.shape[2], tenFlow.shape[3]]
        )
    # end

    tenFlow = torch.cat(
        [
            tenFlow[:, 0:1, :, :] / ((tenInput.shape[3] - 1.0) / 2.0),
            tenFlow[:, 1:2, :, :] / ((tenInput.shape[2] - 1.0) / 2.0),
        ],
        1,
    )
    tenInput = torch.cat([tenInput, backwarp_tenPartial[str(tenFlow.shape)]], 1)

    tenOutput = torch.nn.functional.grid_sample(
        input=tenInput,
        grid=(backwarp_tenGrid[str(tenFlow.shape)] + tenFlow).permute(0, 2, 3, 1),
        mode="bilinear",
        padding_mode="zeros",
        align_corners=False,
    )

    tenMask = tenOutput[:, -1:, :, :]
    tenMask[tenMask > 0.999] = 1.0
    tenMask[tenMask < 1.0] = 0.0

    return tenOutput[:, :-1, :, :] * tenMask


# end

##########################################################


class Network(torch.nn.Module):
    def __init__(self):
        super(Network, self).__init__()

        class Extractor(torch.nn.Module):
            def __init__(self):
                super(Extractor, self).__init__()

                self.netOne = torch.nn.Sequential(
                    torch.nn.Conv2d(
                        in_channels=3,
                        out_channels=16,
                        kernel_size=3,
                        stride=2,
                        padding=1,
                    ),
                    torch.nn.LeakyReLU(inplace=False, negative_slope=0.1),
                    torch.nn.Conv2d(
                        in_channels=16,
                        out_channels=16,
                        kernel_size=3,
                        stride=1,
                        padding=1,
                    ),
                    torch.nn.LeakyReLU(inplace=False, negative_slope=0.1),
                    torch.nn.Conv2d(
                        in_channels=16,
                        out_channels=16,
                        kernel_size=3,
                        stride=1,
                        padding=1,
                    ),
                    torch.nn.LeakyReLU(inplace=False, negative_slope=0.1),
                )

                self.netTwo = torch.nn.Sequential(
                    torch.nn.Conv2d(
                        in_channels=16,
                        out_channels=32,
                        kernel_size=3,
                        stride=2,
                        padding=1,
                    ),
                    torch.nn.LeakyReLU(inplace=False, negative_slope=0.1),
                    torch.nn.Conv2d(
                        in_channels=32,
                        out_channels=32,
                        kernel_size=3,
                        stride=1,
                        padding=1,
                    ),
                    torch.nn.LeakyReLU(inplace=False, negative_slope=0.1),
                    torch.nn.Conv2d(
                        in_channels=32,
                        out_channels=32,
                        kernel_size=3,
                        stride=1,
                        padding=1,
                    ),
                    torch.nn.LeakyReLU(inplace=False, negative_slope=0.1),
                )

                self.netThr = torch.nn.Sequential(
                    torch.nn.Conv2d(
                        in_channels=32,
                        out_channels=64,
                        kernel_size=3,
                        stride=2,
                        padding=1,
                    ),
                    torch.nn.LeakyReLU(inplace=False, negative_slope=0.1),
                    torch.nn.Conv2d(
                        in_channels=64,
                        out_channels=64,
                        kernel_size=3,
                        stride=1,
                        padding=1,
                    ),
                    torch.nn.LeakyReLU(inplace=False, negative_slope=0.1),
                    torch.nn.Conv2d(
                        in_channels=64,
                        out_channels=64,
                        kernel_size=3,
                        stride=1,
                        padding=1,
                    ),
                    torch.nn.LeakyReLU(inplace=False, negative_slope=0.1),
                )

                self.netFou = torch.nn.Sequential(
                    torch.nn.Conv2d(
                        in_channels=64,
                        out_channels=96,
                        kernel_size=3,
                        stride=2,
                        padding=1,
                    ),
                    torch.nn.LeakyReLU(inplace=False, negative_slope=0.1),
                    torch.nn.Conv2d(
                        in_channels=96,
                        out_channels=96,
                        kernel_size=3,
                        stride=1,
                        padding=1,
                    ),
                    torch.nn.LeakyReLU(inplace=False, negative_slope=0.1),
                    torch.nn.Conv2d(
                        in_channels=96,
                        out_channels=96,
                        kernel_size=3,
                        stride=1,
                        padding=1,
                    ),
                    torch.nn.LeakyReLU(inplace=False, negative_slope=0.1),
                )

                self.netFiv = torch.nn.Sequential(
                    torch.nn.Conv2d(
                        in_channels=96,
                        out_channels=128,
                        kernel_size=3,
                        stride=2,
                        padding=1,
                    ),
                    torch.nn.LeakyReLU(inplace=False, negative_slope=0.1),
                    torch.nn.Conv2d(
                        in_channels=128,
                        out_channels=128,
                        kernel_size=3,
                        stride=1,
                        padding=1,
                    ),
                    torch.nn.LeakyReLU(inplace=False, negative_slope=0.1),
                    torch.nn.Conv2d(
                        in_channels=128,
                        out_channels=128,
                        kernel_size=3,
                        stride=1,
                        padding=1,
                    ),
                    torch.nn.LeakyReLU(inplace=False, negative_slope=0.1),
                )

                self.netSix = torch.nn.Sequential(
                    torch.nn.Conv2d(
                        in_channels=128,
                        out_channels=196,
                        kernel_size=3,
                        stride=2,
                        padding=1,
                    ),
                    torch.nn.LeakyReLU(inplace=False, negative_slope=0.1),
                    torch.nn.Conv2d(
                        in_channels=196,
                        out_channels=196,
                        kernel_size=3,
                        stride=1,
                        padding=1,
                    ),
                    torch.nn.LeakyReLU(inplace=False, negative_slope=0.1),
                    torch.nn.Conv2d(
                        in_channels=196,
                        out_channels=196,
                        kernel_size=3,
                        stride=1,
                        padding=1,
                    ),
                    torch.nn.LeakyReLU(inplace=False, negative_slope=0.1),
                )

            # end

            def forward(self, tenInput):
                tenOne = self.netOne(tenInput)
                tenTwo = self.netTwo(tenOne)
                tenThr = self.netThr(tenTwo)
                tenFou = self.netFou(tenThr)
                tenFiv = self.netFiv(tenFou)
                tenSix = self.netSix(tenFiv)

                return [tenOne, tenTwo, tenThr, tenFou, tenFiv, tenSix]

            # end

        # end

        class Decoder(torch.nn.Module):
            def __init__(self, intLevel):
                super(Decoder, self).__init__()

                intPrevious = [
                    None,
                    None,
                    81 + 32 + 2 + 2,
                    81 + 64 + 2 + 2,
                    81 + 96 + 2 + 2,
                    81 + 128 + 2 + 2,
                    81,
                    None,
                ][intLevel + 1]
                intCurrent = [
                    None,
                    None,
                    81 + 32 + 2 + 2,
                    81 + 64 + 2 + 2,
                    81 + 96 + 2 + 2,
                    81 + 128 + 2 + 2,
                    81,
                    None,
                ][intLevel + 0]

                if intLevel < 6:
                    self.netUpflow = torch.nn.ConvTranspose2d(
                        in_channels=2,
                        out_channels=2,
                        kernel_size=4,
                        stride=2,
                        padding=1,
                    )
                if intLevel < 6:
                    self.netUpfeat = torch.nn.ConvTranspose2d(
                        in_channels=intPrevious + 128 + 128 + 96 + 64 + 32,
                        out_channels=2,
                        kernel_size=4,
                        stride=2,
                        padding=1,
                    )
                if intLevel < 6:
                    self.fltBackwarp = [None, None, None, 5.0, 2.5, 1.25, 0.625, None][
                        intLevel + 1
                    ]

                self.netOne = torch.nn.Sequential(
                    torch.nn.Conv2d(
                        in_channels=intCurrent,
                        out_channels=128,
                        kernel_size=3,
                        stride=1,
                        padding=1,
                    ),
                    torch.nn.LeakyReLU(inplace=False, negative_slope=0.1),
                )

                self.netTwo = torch.nn.Sequential(
                    torch.nn.Conv2d(
                        in_channels=intCurrent + 128,
                        out_channels=128,
                        kernel_size=3,
                        stride=1,
                        padding=1,
                    ),
                    torch.nn.LeakyReLU(inplace=False, negative_slope=0.1),
                )

                self.netThr = torch.nn.Sequential(
                    torch.nn.Conv2d(
                        in_channels=intCurrent + 128 + 128,
                        out_channels=96,
                        kernel_size=3,
                        stride=1,
                        padding=1,
                    ),
                    torch.nn.LeakyReLU(inplace=False, negative_slope=0.1),
                )

                self.netFou = torch.nn.Sequential(
                    torch.nn.Conv2d(
                        in_channels=intCurrent + 128 + 128 + 96,
                        out_channels=64,
                        kernel_size=3,
                        stride=1,
                        padding=1,
                    ),
                    torch.nn.LeakyReLU(inplace=False, negative_slope=0.1),
                )

                self.netFiv = torch.nn.Sequential(
                    torch.nn.Conv2d(
                        in_channels=intCurrent + 128 + 128 + 96 + 64,
                        out_channels=32,
                        kernel_size=3,
                        stride=1,
                        padding=1,
                    ),
                    torch.nn.LeakyReLU(inplace=False, negative_slope=0.1),
                )

                self.netSix = torch.nn.Sequential(
                    torch.nn.Conv2d(
                        in_channels=intCurrent + 128 + 128 + 96 + 64 + 32,
                        out_channels=2,
                        kernel_size=3,
                        stride=1,
                        padding=1,
                    )
                )

            # end

            def forward(self, tenFirst, tenSecond, objPrevious):
                tenFlow = None
                tenFeat = None

                if objPrevious is None:
                    tenFlow = None
                    tenFeat = None

                    tenVolume = torch.nn.functional.leaky_relu(
                        input=correlation.FunctionCorrelation(
                            tenFirst=tenFirst, tenSecond=tenSecond
                        ),
                        negative_slope=0.1,
                        inplace=False,
                    )

                    tenFeat = torch.cat([tenVolume], 1)

                elif objPrevious is not None:
                    tenFlow = self.netUpflow(objPrevious["tenFlow"])
                    tenFeat = self.netUpfeat(objPrevious["tenFeat"])

                    tenVolume = torch.nn.functional.leaky_relu(
                        input=correlation.FunctionCorrelation(
                            tenFirst=tenFirst,
                            tenSecond=backwarp(
                                tenInput=tenSecond, tenFlow=tenFlow * self.fltBackwarp
                            ),
                        ),
                        negative_slope=0.1,
                        inplace=False,
                    )

                    tenFeat = torch.cat([tenVolume, tenFirst, tenFlow, tenFeat], 1)

                # end

                tenFeat = torch.cat([self.netOne(tenFeat), tenFeat], 1)
                tenFeat = torch.cat([self.netTwo(tenFeat), tenFeat], 1)
                tenFeat = torch.cat([self.netThr(tenFeat), tenFeat], 1)
                tenFeat = torch.cat([self.netFou(tenFeat), tenFeat], 1)
                tenFeat = torch.cat([self.netFiv(tenFeat), tenFeat], 1)

                tenFlow = self.netSix(tenFeat)

                return {"tenFlow": tenFlow, "tenFeat": tenFeat}

            # end

        # end

        class Refiner(torch.nn.Module):
            def __init__(self):
                super(Refiner, self).__init__()

                self.netMain = torch.nn.Sequential(
                    torch.nn.Conv2d(
                        in_channels=81 + 32 + 2 + 2 + 128 + 128 + 96 + 64 + 32,
                        out_channels=128,
                        kernel_size=3,
                        stride=1,
                        padding=1,
                        dilation=1,
                    ),
                    torch.nn.LeakyReLU(inplace=False, negative_slope=0.1),
                    torch.nn.Conv2d(
                        in_channels=128,
                        out_channels=128,
                        kernel_size=3,
                        stride=1,
                        padding=2,
                        dilation=2,
                    ),
                    torch.nn.LeakyReLU(inplace=False, negative_slope=0.1),
                    torch.nn.Conv2d(
                        in_channels=128,
                        out_channels=128,
                        kernel_size=3,
                        stride=1,
                        padding=4,
                        dilation=4,
                    ),
                    torch.nn.LeakyReLU(inplace=False, negative_slope=0.1),
                    torch.nn.Conv2d(
                        in_channels=128,
                        out_channels=96,
                        kernel_size=3,
                        stride=1,
                        padding=8,
                        dilation=8,
                    ),
                    torch.nn.LeakyReLU(inplace=False, negative_slope=0.1),
                    torch.nn.Conv2d(
                        in_channels=96,
                        out_channels=64,
                        kernel_size=3,
                        stride=1,
                        padding=16,
                        dilation=16,
                    ),
                    torch.nn.LeakyReLU(inplace=False, negative_slope=0.1),
                    torch.nn.Conv2d(
                        in_channels=64,
                        out_channels=32,
                        kernel_size=3,
                        stride=1,
                        padding=1,
                        dilation=1,
                    ),
                    torch.nn.LeakyReLU(inplace=False, negative_slope=0.1),
                    torch.nn.Conv2d(
                        in_channels=32,
                        out_channels=2,
                        kernel_size=3,
                        stride=1,
                        padding=1,
                        dilation=1,
                    ),
                )

            # end

            def forward(self, tenInput):
                return self.netMain(tenInput)

            # end

        # end

        self.netExtractor = Extractor()

        self.netTwo = Decoder(2)
        self.netThr = Decoder(3)
        self.netFou = Decoder(4)
        self.netFiv = Decoder(5)
        self.netSix = Decoder(6)

        self.netRefiner = Refiner()

        self.load_state_dict(
            {
                strKey.replace("module", "net"): tenWeight
                for strKey, tenWeight in torch.hub.load_state_dict_from_url(
                    url="http://content.sniklaus.com/github/pytorch-pwc/network-"
                    + arguments_strModel
                    + ".pytorch",
                    file_name="pwc-" + arguments_strModel,
                ).items()
            }
        )

    # end

    def forward(self, tenFirst, tenSecond):
        tenFirst = self.netExtractor(tenFirst)
        tenSecond = self.netExtractor(tenSecond)

        objEstimate = self.netSix(tenFirst[-1], tenSecond[-1], None)
        objEstimate = self.netFiv(tenFirst[-2], tenSecond[-2], objEstimate)
        objEstimate = self.netFou(tenFirst[-3], tenSecond[-3], objEstimate)
        objEstimate = self.netThr(tenFirst[-4], tenSecond[-4], objEstimate)
        objEstimate = self.netTwo(tenFirst[-5], tenSecond[-5], objEstimate)

        return objEstimate["tenFlow"] + self.netRefiner(objEstimate["tenFeat"])

    # end


# end

netNetwork = None

##########################################################


def estimate(tenFirst, tenSecond):
    global netNetwork

    if netNetwork is None:
        netNetwork = Network().cuda().eval()
    # end

    assert tenFirst.shape[1] == tenSecond.shape[1]
    assert tenFirst.shape[2] == tenSecond.shape[2]

    intWidth = tenFirst.shape[2]
    intHeight = tenFirst.shape[1]

    # assert(intWidth == 1024) # remember that there is no guarantee for correctness, comment this line out if you acknowledge this and want to continue
    # assert(intHeight == 436) # remember that there is no guarantee for correctness, comment this line out if you acknowledge this and want to continue

    tenPreprocessedFirst = tenFirst.cuda().view(1, 3, intHeight, intWidth)
    tenPreprocessedSecond = tenSecond.cuda().view(1, 3, intHeight, intWidth)

    intPreprocessedWidth = int(math.floor(math.ceil(intWidth / 64.0) * 64.0))
    intPreprocessedHeight = int(math.floor(math.ceil(intHeight / 64.0) * 64.0))

    tenPreprocessedFirst = torch.nn.functional.interpolate(
        input=tenPreprocessedFirst,
        size=(intPreprocessedHeight, intPreprocessedWidth),
        mode="bilinear",
        align_corners=False,
    )
    tenPreprocessedSecond = torch.nn.functional.interpolate(
        input=tenPreprocessedSecond,
        size=(intPreprocessedHeight, intPreprocessedWidth),
        mode="bilinear",
        align_corners=False,
    )

    tenFlow = 20.0 * torch.nn.functional.interpolate(
        input=netNetwork(tenPreprocessedFirst, tenPreprocessedSecond),
        size=(intHeight, intWidth),
        mode="bilinear",
        align_corners=False,
    )

    tenFlow[:, 0, :, :] *= float(intWidth) / float(intPreprocessedWidth)
    tenFlow[:, 1, :, :] *= float(intHeight) / float(intPreprocessedHeight)

    return tenFlow[0, :, :, :].cpu()


# end

##########################################################

if __name__ == "__main__":
    if len(sys.argv) < 3:
        raise RuntimeError("Usage: run_batch.py <first_image> <second_image>")
    file1, file2 = sys.argv[1], sys.argv[2]

    tenFirst = torch.FloatTensor(
        numpy.ascontiguousarray(
            numpy.array(PIL.Image.open(file1))[:, :, ::-1]
            .transpose(2, 0, 1)
            .astype(numpy.float32)
            * (1.0 / 255.0)
        )
    )
    tenSecond = torch.FloatTensor(
        numpy.ascontiguousarray(
            numpy.array(PIL.Image.open(file2))[:, :, ::-1]
            .transpose(2, 0, 1)
            .astype(numpy.float32)
            * (1.0 / 255.0)
        )
    )

    tenOutput = estimate(tenFirst, tenSecond)

    with open(arguments_strOut, "wb") as f:
        np.save(f, np.array(tenOutput.numpy().transpose(1, 2, 0), numpy.float32))



import time
import torch
import numpy as np
from neurovc.contrib.flowmag.flow_models.ARFlow.trainer.base_trainer import (
    BaseTrainer,
)
from utils.flow_utils import load_flow, evaluate_flow
from utils.misc_utils import AverageMeter


class TrainFramework(BaseTrainer):
    def __init__(
        self, train_loader, valid_loader, model, loss_func, _log, save_root, config
    ):
        super(TrainFramework, self).__init__(
            train_loader, valid_loader, model, loss_func, _log, save_root, config
        )

    def _run_one_epoch(self):
        am_batch_time = AverageMeter()
        am_data_time = AverageMeter()

        key_meter_names = ["Loss", "l_ph", "l_sm", "flow_mean"]
        key_meters = AverageMeter(i=len(key_meter_names), precision=4)

        self.model.train()
        end = time.time()

        if "stage1" in self.cfg:
            if self.i_epoch == self.cfg.stage1.epoch:
                self.loss_func.cfg.update(self.cfg.stage1.loss)

        for i_step, data in enumerate(self.train_loader):
            if i_step > self.cfg.epoch_size:
                break
            # read data to device
            img1, img2 = data["img1"], data["img2"]
            img_pair = torch.cat([img1, img2], 1).to(self.device)

            # measure data loading time
            am_data_time.update(time.time() - end)

            # compute output
            res_dict = self.model(img_pair, with_bk=True)
            flows_12, flows_21 = res_dict["flows_fw"], res_dict["flows_bw"]
            flows = [
                torch.cat([flo12, flo21], 1) for flo12, flo21 in zip(flows_12, flows_21)
            ]
            loss, l_ph, l_sm, flow_mean = self.loss_func(flows, img_pair)

            # update meters
            key_meters.update(
                [loss.item(), l_ph.item(), l_sm.item(), flow_mean.item()],
                img_pair.size(0),
            )

            # compute gradient and do optimization step
            self.optimizer.zero_grad()
            # loss.backward()

            scaled_loss = 1024.0 * loss
            scaled_loss.backward()

            for param in [p for p in self.model.parameters() if p.requires_grad]:
                param.grad.data.mul_(1.0 / 1024)

            self.optimizer.step()

            # measure elapsed time
            am_batch_time.update(time.time() - end)
            end = time.time()

            if self.i_iter % self.cfg.record_freq == 0:
                for v, name in zip(key_meters.val, key_meter_names):
                    self.summary_writer.add_scalar("Train_" + name, v, self.i_iter)

            if self.i_iter % self.cfg.print_freq == 0:
                istr = (
                    "{}:{:04d}/{:04d}".format(self.i_epoch, i_step, self.cfg.epoch_size)
                    + " Time {} Data {}".format(am_batch_time, am_data_time)
                    + " Info {}".format(key_meters)
                )
                self._log.info(istr)

            self.i_iter += 1
        self.i_epoch += 1

    @torch.no_grad()
    def _validate_with_gt(self):
        batch_time = AverageMeter()

        if type(self.valid_loader) is not list:
            self.valid_loader = [self.valid_loader]

        # only use the first GPU to run validation, multiple GPUs might raise error.
        # https://github.com/Eromera/erfnet_pytorch/issues/2#issuecomment-486142360
        self.model = self.model.module
        self.model.eval()

        end = time.time()

        all_error_names = []
        all_error_avgs = []

        n_step = 0
        for i_set, loader in enumerate(self.valid_loader):
            error_names = ["EPE", "E_noc", "E_occ", "F1_all"]
            error_meters = AverageMeter(i=len(error_names))
            for i_step, data in enumerate(loader):
                img1, img2 = data["img1"], data["img2"]
                img_pair = torch.cat([img1, img2], 1).to(self.device)

                res = list(map(load_flow, data["flow_occ"]))
                gt_flows, occ_masks = [r[0] for r in res], [r[1] for r in res]
                res = list(map(load_flow, data["flow_noc"]))
                _, noc_masks = [r[0] for r in res], [r[1] for r in res]

                gt_flows = [
                    np.concatenate([flow, occ_mask, noc_mask], axis=2)
                    for flow, occ_mask, noc_mask in zip(gt_flows, occ_masks, noc_masks)
                ]

                # compute output
                flows = self.model(img_pair)["flows_fw"]
                pred_flows = flows[0].detach().cpu().numpy().transpose([0, 2, 3, 1])

                es = evaluate_flow(gt_flows, pred_flows)
                error_meters.update([err.item() for err in es], img_pair.size(0))

                # measure elapsed time
                batch_time.update(time.time() - end)
                end = time.time()

                if i_step % self.cfg.print_freq == 0 or i_step == len(loader) - 1:
                    self._log.info(
                        "Test: {0}[{1}/{2}]\t Time {3}\t ".format(
                            i_set, i_step, self.cfg.valid_size, batch_time
                        )
                        + " ".join(map("{:.2f}".format, error_meters.avg))
                    )

                if i_step > self.cfg.valid_size:
                    break
            n_step += len(loader)

            # write error to tf board.
            for value, name in zip(error_meters.avg, error_names):
                self.summary_writer.add_scalar(
                    "Valid_{}_{}".format(name, i_set), value, self.i_epoch
                )

            all_error_avgs.extend(error_meters.avg)
            all_error_names.extend(
                ["{}_{}".format(name, i_set) for name in error_names]
            )

        self.model = torch.nn.DataParallel(self.model, device_ids=self.device_ids)
        # In order to reduce the space occupied during debugging,
        # only the model with more than cfg.save_iter iterations will be saved.
        if self.i_iter > self.cfg.save_iter:
            self.save_model(all_error_avgs[0], name="KITTI_Flow")

        return all_error_avgs, all_error_names



# # todo: rewrite
# from IPython import parallel
# import util as ut, os, sys, time
# #from IPython.parallel.util import interactive
# import subprocess, numpy as np

# if 0:
#   import dill
#   from types import FunctionType
#   from IPython.utils.pickleutil import can_map
#   can_map.pop(FunctionType, None)
#   import pickle
#   from IPython.kernel.zmq import serialize
#   serialize.pickle = pickle

# ip_parallel = parallel

# p = None
# part_info = None
# builtin_map = __builtins__['map']

# def relative_module(m):
#   return hasattr(m, '__file__') \
#          and ((not m.__file__.startswith('/')) \
#               or m.__file__.startswith(os.getcwd()))

# def safe_reload(m):
#   if relative_module(m) and (not hasattr(m, '__no_autoreload__')):
#     print 'reloading'
#     reload(m)

# def run_reload():
#   excludes = set(['__main__', 'autoreload'])
#   for name, m in list(sys.modules.iteritems()):
#     if m and relative_module(m) and (name not in excludes) and (not hasattr(m, '__no_autoreload__')):
#       reload(m)

# class Parallel:
#   def __init__(self, config = None, rc = None, view = None, bview = None):
#     self.config = config
#     self._rc = rc
#     self._view = view
#     self._bview = bview

#   def init_rc(self, reset = False):
#     if reset or (self._rc is None):
#       if self.config is None:
#         self._rc = ip_parallel.Client()
#       else:
#         self._rc = ip_parallel.Client(self.config)
#       self._view = self._rc.direct_view()
#       #self._view = self._rc.load_balanced_view()
#       self._bview = self._rc.load_balanced_view()

#   def view(self):
#     self.init_rc()
#     return self._view

#   def bview(self):
#     self.init_rc()
#     return self._bview


#   def rc(self):
#     self.init_rc()
#     return self._rc

#   def per_machine_subset(self, n_per_machine):
#     targets = self.targets_by_hostname()
#     target_subset = ut.flatten([ut.sample_at_most(ts, n_per_machine) \
#                                 for ts in targets.values()])
#     return Parallel(self.config, self.rc(), self.rc()[target_subset])

#   def partition(self, n, i):
#     target_subset = [j for j in self.rc()[:].targets if j % n == i]
#     return Parallel(self.config, self.rc(), self.rc()[target_subset])

#   def targets_by_hostname(self):
#     # initialize parallel_rc
#     self.import_modules(['subprocess'], False)
#     targets_by_hostname = {}
#     for t in self.rc()[:].targets:
#       v = self.rc()[[t]]
#       [hostname] = v.map_sync(interactive(lambda x: subprocess.Popen('hostname', shell=True, stdout=subprocess.PIPE).stdout.read().rstrip()), [1])
#       if hostname not in targets_by_hostname:
#         targets_by_hostname[hostname] = []
#       targets_by_hostname[hostname].append(t)
#     return targets_by_hostname

#   def execute(self, code):
#     r = self.view().execute(code)
#     r.wait()
#     if not r.successful():
#       print 'iputil.execute error:'
#       print r.result()
#     assert r.successful()

#   # def list_machines(self):
#   #   self.view().execute('import subprocess, os').wait()
#   #   res = self.view().map_sync(interactive(lambda x : subprocess.Popen('hostname', shell=True,
#   #                                                                 stdout=subprocess.PIPE).stdout.read().rstrip()), range(200))
#   #   print 'Running on:'
#   #   ut.prn_lines(set(res))

#   def list_machines(self):
#     self.view().execute('import subprocess, os').wait()
#     # res = self.view().map_sync(interactive(lambda x : (subprocess.Popen('hostname', shell=True, stdout=subprocess.PIPE).stdout.read().rstrip(),
#     #                                                    os.getpid())), range(2000))
#     res = self.view().map_sync(lambda x : (subprocess.Popen('hostname', shell=True, stdout=subprocess.PIPE).stdout.read().rstrip(),
#                                os.getpid()), range(2000))
#     print 'Running on:'
#     machines = {}
#     for x, p in res:
#       if x not in machines:
#         machines[x] = set()
#       machines[x].add(p)
#     for x, ps in machines.items():
#       print x, len(ps)

#   def send_var(self, var_name, val):
#     res = self.view().push({var_name : val})
#     res.wait()
#     assert res.successful()

#   def import_modules(self, packages, reload = False):
#     packages = [('aolib.iputil', 'iputil')] + list(packages)
#     import_cmd = ''
#     reload_cmd = ''
#     for p in packages:
#       if len(import_cmd) != 0:
#         import_cmd += ', '
#         reload_cmd += '; '
#       if type(p) == type((1,)):
#         import_cmd += '%s as %s' % p
#         reload_cmd += 'iputil.safe_reload(%s)' % p[1]
#       else:
#         import_cmd += p
#         reload_cmd += 'iputil.safe_reload(%s)' % p

#     if reload:
#       cmd = 'import %s; %s' % (import_cmd, reload_cmd)
#     else:
#       cmd = 'import %s' % import_cmd
#     self.execute(cmd)

#   def reload_modules(self):
#     self.execute('from aolib import iputil as iputil, areload; iputil.run_reload()')
#     #self.execute('import aolib.iputil as iputil, areload; iputil.run_reload()')

#   def send_var_fs(self, var_name, val):
#     fname = ut.make_temp_nfs('.pk')
#     ut.save(fname, val)
#     res = self.view().execute('import util as ut; ut.wait_for_file("%s", 30); %s = ut.load("%s")' % (fname, var_name, fname))
#     res.wait()
#     assert res.successful()
#     os.remove(fname)

#   def clear(self):
#     self.view().results.clear()
#     self.rc().results.clear()
#     self.rc().metadata.clear()

#   # def maybe_map_sync(self, parallel, f, xs, imports = []):
#   #   if parallel:
#   #     return self.map_sync(f, xs, imports)
#   #   else:
#   #     # assume imports have already been loaded if not parallel
#   #     if type(xs) == type((1,)):
#   #       return map(f, *xs)
#   #     else:
#   #       return map(f, xs)

#   # def maybe_map_sync(self, parallel, f, xs, imports = [], vars = {}):
#   #   if parallel:
#   #     return self.map_sync(f, xs, imports)
#   #   else:
#   #     # todo: is it better to save these temporarily? or would it only make it more confusing?
#   #     f.func_globals.update(vars)
#   #     # assume imports have already been loaded if not parallel
#   #     if type(xs) == type((1,)):
#   #       return map(f, *xs)
#   #     else:
#   #       return map(f, xs)

#   def maybe_map_sync(self, parallel, f, xs, vars = {}, imports = [], do_reload = True):
#     if parallel:
#       return self.map_sync(f, xs, vars, imports, do_reload = do_reload)
#     else:
#       # todo: is it better to save these temporarily? or would it only make it more confusing?
#       if len(vars):
#         f.func_globals.update(vars)
#       # assume imports have already been loaded if not parallel
#       if type(xs) == type((1,)):
#         return builtin_map(f, *xs)
#       else:
#         return builtin_map(f, xs)

#   def map(self, parallel, f, xs, vars = {}, imports = [], do_reload = True):
#     return self.maybe_map_sync(parallel, f, xs, vars, imports, do_reload = do_reload)

#   def amap(self, *args, **kwargs):
#     return np.array(self.map(*args, **kwargs))

#   def map_sync(self, f, xs, vars = {}, imports = [], do_reload = True, do_reset = False, use_bview = False):
#     # if do_reset:
#     #   self.init_rc(do_reset)

#     if len(imports):
#       self.import_modules(imports, True)
#     if do_reload:
#       self.reload_modules()

#     if len(xs) == 0:
#       return []

#     if use_bview:
#       view = self.view() if do_reload else self.bview()
#     else:
#       view = self.view()
#     # interpret tuple as multiple args
#     if type(xs) == type((1,)):
#       res = view.map_sync(f, *xs)
#     else:
#       res = view.map_sync(f, xs)

#     self.clear()

#     return res

#   def abort(self):
#     self.view().abort()

# #  def test(self, do = True, max_wait = 30):
#   def test(self, do = True, max_wait = 60):
#     if do:
#       a = self.view().map_async(lambda x: x, range(100))
#       start = time.time()
#       while time.time() - start <= max_wait:
#         if a.ready():
#           return True
#         else:
#           time.sleep(1)
#       raise RuntimeError('test() timed out after %s seconds!' % max_wait)

# __no_autoreload__ = True

# def init(do=True, local = False):
#   if do:
#     global p, part_info
#     if local:
#       config_file = None
#     else:
#       #config_file = '/csail/vision-billf5/aho/ipcluster_profile/security/ipcontroller-client.json'
#       config_file = '/data/vision/billf/aho-billf/ipcluster_profile/security/ipcontroller-client.json'
#     if p is not None and p._rc is not None:
#       p._rc.close()
#     p = Parallel(config_file)
#     if part_info is not None:
#       print 'iputil partition:', part_info
#       p = p.partition(*part_info)
#     for f in dir(p):
#       if not f.startswith('_') and hasattr(getattr(p, f), '__call__'):
#         globals()[f] = getattr(p, f)

# init()

# # def restart():
# #   os.system('./cluster.sh')
# #   init()


# def global_partition(n=None, i=None):
#   global p, part_info
#   if n is None:
#     part_info = None
#   else:
#     part_info = (n, i)
#   init()

# def test_partition():
#   def f(x):
#     import iputil
#     return sum(iputil.partition(2, 1).map_sync(lambda x: x, range(10)))
#   import iputil
#   iputil.map_sync(lambda x:x, range(10000))
#   assert sum(iputil.partition(2, 0).map_sync(f, range(1000))) == 1000*sum(range(10))

# def reset(parallel=True):
#   init(parallel)
#   test(parallel)

# def send(config_file = '/data/vision/billf/aho-billf/ipcluster_profile/security/ipcontroller-client.json'):
#   rc = ip_parallel.Client(config_file)
